#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle
import threading
import time
import traceback
from datetime import datetime
from typing import List, Union, Dict

from fim.slivers.network_link import NetworkLinkSliver

from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin, ActorType
from fabric_cf.actor.core.apis.abc_broker_proxy import ABCBrokerProxy
from fabric_cf.actor.core.apis.abc_controller_reservation import ABCControllerReservation
from fabric_cf.actor.core.apis.abc_database import ABCDatabase
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin, ReservationCategory
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import DatabaseException
from fabric_cf.actor.core.kernel.poa import Poa, PoaStates
from fabric_cf.actor.core.kernel.slice import SliceTypes
from fabric_cf.actor.core.plugins.handlers.configuration_mapping import ConfigurationMapping
from fabric_cf.actor.core.container.maintenance import Site
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.db.psql_database import PsqlDatabase


class ActorDatabase(ABCDatabase):
    MAINTENANCE = 'maintenance'

    def __init__(self, *, user: str, password: str, database: str, db_host: str, logger):
        self.user = user
        self.password = password
        self.database = database
        self.db_host = db_host
        self.db = PsqlDatabase(user=self.user, password=self.password, database=self.database, db_host=self.db_host,
                               logger=logger)
        self.actor_type = None
        self.actor_name = None
        self.actor_id = None
        self.initialized = False
        self.logger = logger
        self.reset_state = False
        self.actor = None
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['db']
        del state['actor_type']
        del state['actor_name']
        del state['actor_id']
        del state['initialized']
        del state['logger']
        del state['reset_state']
        del state['actor']
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.db = PsqlDatabase(user=self.user, password=self.password, database=self.database, db_host=self.db_host,
                               logger=None)
        self.actor_id = None
        self.actor_name = None
        self.actor_type = None
        self.initialized = False
        self.logger = None
        self.reset_state = False
        self.lock = threading.Lock()

    def set_logger(self, *, logger):
        self.logger = logger
        if self.db is not None:
            self.db.set_logger(logger=logger)

    def set_reset_state(self, *, state: bool):
        self.reset_state = state

    def initialize(self):
        if not self.initialized:
            if self.actor_name is None:
                raise DatabaseException(Constants.NOT_SPECIFIED_PREFIX.format("actor name"))
            self.initialized = True

    def actor_added(self, *, actor):
        self.actor = actor
        self.actor_id = self.get_actor_id_from_name(actor_name=self.actor_name)
        if self.actor_id is None:
            raise DatabaseException(Constants.OBJECT_NOT_FOUND.format("actor", self.actor_name))

    def revisit(self, *, actor: ABCActorMixin, properties: dict):
        return

    def get_actor_id_from_name(self, *, actor_name: str) -> int or None:
        try:
            #self.lock.acquire()
            actor = self.db.get_actor(name=actor_name)
            self.actor_id = actor['act_id']
            self.actor_type = ActorType(actor['act_type'])
            return self.actor_id
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return None

    def get_slice_by_id(self, *, slc_id: int) -> ABCSlice or None:
        try:
            #self.lock.acquire()
            slice_dict = self.db.get_slice_by_id(slc_id=slc_id)
            if slice_dict is not None:
                pickled_slice = slice_dict.get(Constants.PROPERTY_PICKLE_PROPERTIES)
                return pickle.loads(pickled_slice)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return None

    def add_slice(self, *, slice_object: ABCSlice):
        try:
            slices = self.get_slices(slice_id=slice_object.get_slice_id())
            if len(slices) > 0:
                raise DatabaseException("Slice # {} already exists".format(slice_object.get_slice_id()))

            properties = pickle.dumps(slice_object)
            oidc_claim_sub = None
            email = None
            if slice_object.get_owner() is not None:
                oidc_claim_sub = slice_object.get_owner().get_oidc_sub_claim()
                email = slice_object.get_owner().get_email()

            #self.lock.acquire()
            self.db.add_slice(slc_guid=str(slice_object.get_slice_id()),
                              slc_state=slice_object.get_state().value,
                              slc_name=slice_object.get_name(),
                              slc_type=slice_object.get_slice_type().value,
                              slc_resource_type=str(slice_object.get_resource_type()),
                              properties=properties,
                              slc_graph_id=slice_object.get_graph_id(),
                              oidc_claim_sub=oidc_claim_sub, email=email,
                              lease_end=slice_object.get_lease_end(),
                              lease_start=slice_object.get_lease_start(),
                              project_id=slice_object.get_project_id())
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_slice(self, *, slice_object: ABCSlice):
        # Update the slice only when there are changes to be reflected in database
        if not slice_object.is_dirty():
            return
        slice_object.clear_dirty()
        try:
            #self.lock.acquire()
            properties = pickle.dumps(slice_object)
            self.db.update_slice(slc_guid=str(slice_object.get_slice_id()),
                                 slc_name=slice_object.get_name(),
                                 slc_type=slice_object.get_slice_type().value,
                                 slc_state=slice_object.get_state().value,
                                 slc_resource_type=str(slice_object.get_resource_type()),
                                 properties=properties,
                                 slc_graph_id=slice_object.get_graph_id(),
                                 lease_end=slice_object.get_lease_end(), lease_start=slice_object.get_lease_start())
        finally:
            if self.lock.locked():
                self.lock.release()

    def remove_slice(self, *, slice_id: ID):
        try:
            #self.lock.acquire()
            self.db.remove_slice(slc_guid=str(slice_id))
        finally:
            if self.lock.locked():
                self.lock.release()

    def get_slices(self, *, slice_id: ID = None, slice_name: str = None, project_id: str = None, email: str = None,
                   states: list[int] = None, oidc_sub: str = None, slc_type: List[SliceTypes] = None,
                   limit: int = None, offset: int = None, lease_end: datetime = None,
                   search: str = None, exact_match: bool = False,
                   updated_after: datetime = None) -> List[ABCSlice] or None:
        result = []
        try:
            try:
                #self.lock.acquire()
                slice_type = None
                if slc_type is not None:
                    slice_type = [x.value for x in slc_type]
                sid = str(slice_id) if slice_id is not None else None
                slices = self.db.get_slices(slice_id=sid, slice_name=slice_name, project_id=project_id, email=email,
                                            states=states, oidc_sub=oidc_sub, slc_type=slice_type, limit=limit,
                                            offset=offset, lease_end=lease_end, search=search, exact_match=exact_match,
                                            updated_after=updated_after)
            finally:
                if self.lock.locked():
                    self.lock.release()
            if slices is not None:
                for s in slices:
                    pickled_slice = s.get(Constants.PROPERTY_PICKLE_PROPERTIES)
                    slice_obj = pickle.loads(pickled_slice)
                    slice_obj.set_last_updated_time(s.get('last_updated_time'))
                    result.append(slice_obj)
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def increment_metrics(self, *, project_id: str, oidc_sub: str, slice_count: int = 1) -> bool:
        try:
            self.lock.acquire()
            self.db.increment_metrics(project_id=project_id, user_id=oidc_sub, slice_count=slice_count)
            return True
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        return False

    def get_metrics(self, *, project_id: str, oidc_sub: str, excluded_projects: List[str] = None) -> list:
        try:
            return self.db.get_metrics(project_id=project_id, user_id=oidc_sub, excluded_projects=excluded_projects)
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()

    def get_slice_count(self, *, project_id: str = None, email: str = None, states: list[int] = None,
                        oidc_sub: str = None, slc_type: List[SliceTypes] = None,
                        excluded_projects: List[str] = None) -> int:
        try:
            slice_type = [SliceTypes.ClientSlice.value]
            if slc_type is not None:
                slice_type = [x.value for x in slc_type]
            return self.db.get_slice_count(project_id=project_id, email=email, states=states, oidc_sub=oidc_sub,
                                           slc_type=slice_type, excluded_projects=excluded_projects)
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        return -1

    def add_reservation(self, *, reservation: ABCReservationMixin):
        try:
            #self.lock.acquire()
            self.logger.debug("Adding reservation {} to slice {}".format(reservation.get_reservation_id(),
                                                                         reservation.get_slice()))
            properties = pickle.dumps(reservation)
            oidc_claim_sub = None
            email = None
            if reservation.get_slice() is not None and reservation.get_slice().get_owner() is not None:
                oidc_claim_sub = reservation.get_slice().get_owner().get_oidc_sub_claim()
                email = reservation.get_slice().get_owner().get_email()

            site = None
            rsv_type = None
            components = None
            host = None
            ip_subnet = None
            sliver = None
            links = []
            from fabric_cf.actor.core.kernel.reservation_client import ReservationClient
            if isinstance(reservation, ReservationClient) and reservation.get_leased_resources() and \
                    reservation.get_leased_resources().get_sliver():
                sliver = reservation.get_leased_resources().get_sliver()
            if not sliver and reservation.get_resources() and reservation.get_resources().get_sliver():
                sliver = reservation.get_resources().get_sliver()

            if sliver:
                rsv_type = sliver.get_type().name
                from fim.slivers.network_service import NetworkServiceSliver
                from fim.slivers.network_node import NodeSliver

                if isinstance(sliver, NetworkServiceSliver) and sliver.interface_info:
                    site = sliver.get_site()
                    if sliver.get_gateway():
                        ip_subnet = sliver.get_gateway().subnet

                    components = []
                    for interface in sliver.interface_info.interfaces.values():
                        graph_id_node_id_component_id, bqm_if_name = interface.get_node_map()
                        if ":" in graph_id_node_id_component_id or "#" in graph_id_node_id_component_id:
                            if "#" in graph_id_node_id_component_id:
                                split_string = graph_id_node_id_component_id.split("#")
                            else:
                                split_string = graph_id_node_id_component_id.split(":")
                            node_id = split_string[1] if len(split_string) > 1 else None
                            comp_id = split_string[2] if len(split_string) > 2 else None
                            bdf = ":".join(split_string[3:]) if len(split_string) > 3 else None
                            if node_id and comp_id and bdf:
                                components.append((node_id, comp_id, bdf))
                    if sliver.ero and sliver.capacities:
                        type, path = sliver.ero.get()
                        if path and len(path.get()):
                            for hop in path.get()[0]:
                                if hop.startswith('link:'):
                                    links.append({"node_id": hop,
                                                  "bw": sliver.capacities.bw})

                elif isinstance(sliver, NodeSliver):
                    site = sliver.get_site()
                    if sliver.get_labels() and sliver.get_labels().instance_parent:
                        host = sliver.get_labels().instance_parent
                    if sliver.get_label_allocations() and sliver.get_label_allocations().instance_parent:
                        host = sliver.get_label_allocations().instance_parent
                    if sliver.get_management_ip():
                        ip_subnet = str(sliver.get_management_ip())

                    node_id = reservation.get_graph_node_id()
                    if node_id and sliver.attached_components_info:
                        components = []
                        for c in sliver.attached_components_info.devices.values():
                            if c.get_node_map():
                                bqm_id, comp_id = c.get_node_map()
                                if c.labels and c.labels.bdf:
                                    bdf = c.labels.bdf
                                    if isinstance(c.labels.bdf, str):
                                        bdf = [c.labels.bdf]
                                    for x in bdf:
                                        components.append((node_id, comp_id, x))

            term = reservation.get_term()

            self.db.add_reservation(slc_guid=str(reservation.get_slice_id()),
                                    rsv_resid=str(reservation.get_reservation_id()),
                                    rsv_category=reservation.get_category().value,
                                    rsv_state=reservation.get_state().value,
                                    rsv_pending=reservation.get_pending_state().value,
                                    rsv_joining=reservation.get_join_state().value,
                                    properties=properties,
                                    rsv_graph_node_id=reservation.get_graph_node_id(),
                                    oidc_claim_sub=oidc_claim_sub, email=email, site=site, rsv_type=rsv_type,
                                    components=components,
                                    lease_start=term.get_start_time() if term else None,
                                    lease_end=term.get_end_time() if term else None,
                                    host=host, ip_subnet=ip_subnet, links=links)
            self.logger.debug(
                "Reservation {} added to slice {}".format(reservation.get_reservation_id(), reservation.get_slice()))
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_reservation(self, *, reservation: ABCReservationMixin):
        # Update the reservation only when there are changes to be reflected in database
        if not reservation.is_dirty():
            return
        reservation.clear_dirty()
        try:
            #self.lock.acquire()
            self.logger.debug("Updating reservation {} in slice {}".format(reservation.get_reservation_id(),
                                                                           reservation.get_slice()))

            site = None
            rsv_type = None
            components = None
            ip_subnet = None
            host = None
            sliver = None
            links = []
            from fabric_cf.actor.core.kernel.reservation_client import ReservationClient
            if isinstance(reservation, ReservationClient) and reservation.get_leased_resources() and \
                    reservation.get_leased_resources().get_sliver():
                sliver = reservation.get_leased_resources().get_sliver()
            if not sliver and reservation.get_resources() and reservation.get_resources().get_sliver():
                sliver = reservation.get_resources().get_sliver()

            if sliver:
                rsv_type = sliver.get_type().name
                from fim.slivers.network_service import NetworkServiceSliver
                from fim.slivers.network_node import NodeSliver
                if isinstance(sliver, NetworkServiceSliver) and sliver.interface_info:
                    site = sliver.get_site()

                    if sliver.get_gateway():
                        ip_subnet = sliver.get_gateway().subnet

                    components = []
                    for interface in sliver.interface_info.interfaces.values():
                        graph_id_node_id_component_id, bqm_if_name = interface.get_node_map()
                        if ":" in graph_id_node_id_component_id or "#" in graph_id_node_id_component_id:
                            if "#" in graph_id_node_id_component_id:
                                split_string = graph_id_node_id_component_id.split("#")
                            else:
                                split_string = graph_id_node_id_component_id.split(":")
                            node_id = split_string[1] if len(split_string) > 1 else None
                            comp_id = split_string[2] if len(split_string) > 2 else None
                            bdf = ":".join(split_string[3:]) if len(split_string) > 3 else None
                            if node_id and comp_id and bdf:
                                components.append((node_id, comp_id, bdf))

                    if sliver.ero and sliver.capacities:
                        type, path = sliver.ero.get()
                        if path and len(path.get()):
                            for hop in path.get()[0]:
                                if hop.startswith('link:'):
                                    links.append({"node_id": hop,
                                                  "bw": sliver.capacities.bw})
                elif isinstance(sliver, NodeSliver):
                    site = sliver.get_site()

                    if sliver.get_labels() and sliver.get_labels().instance_parent:
                        host = sliver.get_labels().instance_parent
                    if sliver.get_label_allocations() and sliver.get_label_allocations().instance_parent:
                        host = sliver.get_label_allocations().instance_parent
                    if sliver.get_management_ip():
                        ip_subnet = str(sliver.get_management_ip())
                    node_id = reservation.get_graph_node_id()
                    if node_id and sliver.attached_components_info:
                        components = []
                        for c in sliver.attached_components_info.devices.values():
                            if c.get_node_map():
                                bqm_id, comp_id = c.get_node_map()
                                if c.labels and c.labels.bdf:
                                    bdf = c.labels.bdf
                                    if isinstance(c.labels.bdf, str):
                                        bdf = [c.labels.bdf]
                                    for x in bdf:
                                        components.append((node_id, comp_id, x))

            term = reservation.get_term()
            begin = time.time()
            properties = pickle.dumps(reservation)
            diff = int(time.time() - begin)
            if diff > 0:
                self.logger.info(f"PICKLE TIME: {diff}")
            begin = time.time()
            self.db.update_reservation(slc_guid=str(reservation.get_slice_id()),
                                       rsv_resid=str(reservation.get_reservation_id()),
                                       rsv_category=reservation.get_category().value,
                                       rsv_state=reservation.get_state().value,
                                       rsv_pending=reservation.get_pending_state().value,
                                       rsv_joining=reservation.get_join_state().value,
                                       properties=properties,
                                       rsv_graph_node_id=reservation.get_graph_node_id(),
                                       site=site, rsv_type=rsv_type, components=components,
                                       lease_start=term.get_start_time() if term else None,
                                       lease_end=term.get_end_time() if term else None,
                                       ip_subnet=ip_subnet, host=host, links=links)
            diff = int(time.time() - begin)
            if diff > 0:
                self.logger.info(f"DB TIME: {diff}")
        finally:
            if self.lock.locked():
                self.lock.release()

    def remove_reservation(self, *, rid: ID):
        try:
            #self.lock.acquire()
            self.logger.debug("Removing reservation {}".format(rid))
            try:
                self.db.remove_unit(unt_uid=str(rid))
            except Exception as e:
                self.logger.debug("No associated Unit associated with the reservation")
            self.db.remove_reservation(rsv_resid=str(rid))
        finally:
            if self.lock.locked():
                self.lock.release()

    def _load_reservation_from_pickled_object(self, pickled_res: bytes, slc_id: int) -> ABCReservationMixin or None:
        try:
            slice_obj = self.get_slice_by_id(slc_id=slc_id)
            result = pickle.loads(pickled_res)
            result.restore(actor=self.actor, slice_obj=slice_obj)

            if isinstance(result, ABCControllerReservation):
                if result.get_redeem_predecessors() is not None:
                    for p in result.get_redeem_predecessors():
                        if p.reservation_id is not None:
                            parent = self.get_reservations(rid=p.reservation_id)
                            if parent is not None and len(parent) > 0:
                                p.set_reservation(reservation=parent[0])

                    for p in result.get_join_predecessors():
                        if p.reservation_id is not None:
                            parent = self.get_reservations(rid=p.reservation_id)
                            if parent is not None and len(parent) > 0:
                                p.set_reservation(reservation=parent[0])

            # Load in progress POAs
            poa_list = self.get_poas(sliver_id=result.get_reservation_id(), include_res_info=False,
                                     states=[PoaStates.Nascent.value, PoaStates.Performing.value,
                                             PoaStates.AwaitingCompletion.value, PoaStates.SentToAuthority.value])

            from fabric_cf.actor.core.kernel.reservation_client import ReservationClient
            from fabric_cf.actor.core.kernel.authority_reservation import AuthorityReservation
            if isinstance(result, ReservationClient) or isinstance(result, AuthorityReservation):
                for poa in poa_list:
                    poa.restore(actor=self.actor, reservation=result)
                    result.poas[poa.get_poa_id()] = poa

            return result
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        return None

    def _load_reservations_from_db(self, *, res_dict_list: List[dict]) -> List[ABCReservationMixin]:
        result = []
        if res_dict_list is None:
            return result

        for r in res_dict_list:
            pickled_res = r.get(Constants.PROPERTY_PICKLE_PROPERTIES)
            slc_id = r.get(Constants.RSV_SLC_ID)
            res_obj = self._load_reservation_from_pickled_object(pickled_res=pickled_res, slc_id=slc_id)
            result.append(res_obj)
        return result

    def get_client_reservations(self, *, slice_id: ID = None) -> List[ABCReservationMixin]:
        result = []
        try:
            #self.lock.acquire()
            sid = str(slice_id) if slice_id is not None else None
            res_dict_list = self.db.get_reservations(slice_id=sid,
                                                     category=[ReservationCategory.Broker.value,
                                                               ReservationCategory.Authority.value])
            if self.lock.locked():
                self.lock.release()
            result = self._load_reservations_from_db(res_dict_list=res_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def get_holdings(self, *, slice_id: ID = None) -> List[ABCReservationMixin]:
        result = []
        try:
            #self.lock.acquire()
            sid = str(slice_id) if slice_id is not None else None
            res_dict_list = self.db.get_reservations(slice_id=sid, category=[ReservationCategory.Client.value])
            if self.lock.locked():
                self.lock.release()
            result = self._load_reservations_from_db(res_dict_list=res_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def get_broker_reservations(self) -> List[ABCReservationMixin]:
        result = []
        try:
            #self.lock.acquire()
            res_dict_list = self.db.get_reservations(category=[ReservationCategory.Broker.value])
            if self.lock.locked():
                self.lock.release()
            result = self._load_reservations_from_db(res_dict_list=res_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def get_authority_reservations(self) -> List[ABCReservationMixin]:
        result = []
        try:
            #self.lock.acquire()
            result = []
            res_dict_list = self.db.get_reservations(category=[ReservationCategory.Authority.value])
            if self.lock.locked():
                self.lock.release()
            result = self._load_reservations_from_db(res_dict_list=res_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def get_components(self, *, node_id: str, states: list[int], rsv_type: list[str], component: str = None,
                       bdf: str = None, start: datetime = None, end: datetime = None,
                       excludes: List[str] = None) -> Dict[str, List[str]]:
        try:
            return self.db.get_components(node_id=node_id, states=states, component=component, bdf=bdf,
                                          rsv_type=rsv_type, start=start, end=end, excludes=excludes)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()

    def get_links(self, *, node_id: str, states: list[int], rsv_type: list[str], start: datetime = None,
                  end: datetime = None, excludes: List[str] = None) -> Dict[str, int]:
        try:
            return self.db.get_links(node_id=node_id, states=states, rsv_type=rsv_type, start=start,
                                     end=end, excludes=excludes)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()

    def get_reservations(self, *, slice_id: ID = None, graph_node_id: str = None, project_id: str = None,
                         email: str = None, oidc_sub: str = None, rid: ID = None, states: list[int] = None,
                         site: str = None, rsv_type: list[str] = None, start: datetime = None,
                         end: datetime = None, ip_subnet: str = None, host: str = None) -> List[ABCReservationMixin]:
        result = []
        try:
            #self.lock.acquire()
            sid = str(slice_id) if slice_id is not None else None
            res_id = str(rid) if rid is not None else None
            res_dict_list = self.db.get_reservations(slice_id=sid, graph_node_id=graph_node_id, host=host, ip_subnet=ip_subnet,
                                                     project_id=project_id, email=email, oidc_sub=oidc_sub, rid=res_id,
                                                     states=states, site=site, rsv_type=rsv_type, start=start, end=end)
            if self.lock.locked():
               self.lock.release()
            result = self._load_reservations_from_db(res_dict_list=res_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def get_reservations_by_rids(self, *, rid: List[str]) -> List[ABCReservationMixin]:
        result = []
        try:
            #self.lock.acquire()
            res_dict_list = self.db.get_reservations_by_rids(rsv_resid_list=rid)
            if self.lock.locked():
                self.lock.release()
            result = self._load_reservations_from_db(res_dict_list=res_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def add_broker(self, *, broker: ABCBrokerProxy):
        try:
            #self.lock.acquire()
            self.logger.debug("Adding broker {}({})".format(broker.get_name(), broker.get_guid()))
            properties = pickle.dumps(broker)
            self.db.add_proxy(act_id=self.actor_id, prx_name=broker.get_name(), properties=properties)
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_broker(self, *, broker: ABCBrokerProxy):
        try:
            #self.lock.acquire()
            self.logger.debug("Updating broker {}({})".format(broker.get_name(), broker.get_guid()))
            properties = pickle.dumps(broker)
            self.db.update_proxy(act_id=self.actor_id, prx_name=broker.get_name(), properties=properties)
        finally:
            if self.lock.locked():
                self.lock.release()

    def remove_broker(self, *, broker: ABCBrokerProxy):
        try:
            #self.lock.acquire()
            self.logger.debug("Removing broker {}({})".format(broker.get_name(), broker.get_guid()))
            self.db.remove_proxy(act_id=self.actor_id, prx_name=broker.get_name())
        finally:
            if self.lock.locked():
                self.lock.release()

    def set_actor_name(self, *, name: str):
        self.actor_name = name

    def get_brokers(self) -> List[ABCBrokerProxy] or None:
        try:
            #self.lock.acquire()
            result = []
            broker_dict_list = self.db.get_proxies(act_id=self.actor_id)
            if broker_dict_list is not None:
                for b in broker_dict_list:
                    pickled_broker = b.get(Constants.PROPERTY_PICKLE_PROPERTIES)
                    broker_obj = pickle.loads(pickled_broker)
                    result.append(broker_obj)
            return result
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        return None

    def add_delegation(self, *, delegation: ABCDelegation):
        self.logger.debug("Adding delegation {} to slice {}".format(delegation.get_delegation_id(),
                                                                    delegation.get_slice_id()))
        try:
            #self.lock.acquire()
            properties = pickle.dumps(delegation)
            self.db.add_delegation(slice_id=str(delegation.get_slice_id()),
                                   dlg_graph_id=str(delegation.get_delegation_id()),
                                   dlg_state=delegation.get_state().value,
                                   properties=properties,
                                   site=delegation.get_site())
            self.logger.debug(
                "Delegation {} added to slice {}".format(delegation.get_delegation_id(),
                                                         delegation.get_slice_id()))
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_delegation(self, *, delegation: ABCDelegation):
        # Update the delegation only when there are changes to be reflected in database
        if not delegation.is_dirty():
            return
        delegation.clear_dirty()
        try:
            #self.lock.acquire()
            self.logger.debug("Updating delegation {} in slice {}".format(delegation.get_delegation_id(),
                                                                          delegation.get_slice_id()))
            properties = pickle.dumps(delegation)
            self.db.update_delegation(dlg_graph_id=str(delegation.get_delegation_id()),
                                      dlg_state=delegation.get_state().value,
                                      properties=properties,
                                      site=delegation.get_site())
        finally:
            if self.lock.locked():
                self.lock.release()

    def remove_delegation(self, *, dlg_graph_id: str):
        try:
            #self.lock.acquire()
            self.logger.debug("Removing delegation {}".format(dlg_graph_id))
            self.db.remove_delegation(dlg_graph_id=str(dlg_graph_id))
        finally:
            if self.lock.locked():
                self.lock.release()

    def _load_delegation_from_pickled_instance(self, pickled_del: bytes, slc_id: int) -> ABCDelegation:
        slice_obj = self.get_slice_by_id(slc_id=slc_id)
        delegation = pickle.loads(pickled_del)
        delegation.restore(actor=self.actor, slice_obj=slice_obj)
        return delegation

    def _load_delegation_from_db(self, dlg_dict_list: List[dict]) -> List[ABCDelegation]:
        result = []
        if dlg_dict_list is None:
            return result

        for d in dlg_dict_list:
            pickled_del = d.get(Constants.PROPERTY_PICKLE_PROPERTIES)
            slc_id = d.get(Constants.DLG_SLC_ID)
            dlg_obj = self._load_delegation_from_pickled_instance(pickled_del=pickled_del, slc_id=slc_id)
            result.append(dlg_obj)

        return result

    def get_delegation(self, *, dlg_graph_id: str) -> ABCDelegation or None:
        try:
            #self.lock.acquire()
            dlg_dict = self.db.get_delegation(dlg_graph_id=str(dlg_graph_id))
            if self.lock.locked():
                self.lock.release()
            return self._load_delegation_from_db(dlg_dict_list=[dlg_dict])[0]
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return None

    def get_delegations(self, *, slice_id: ID = None, states: List[int] = None) -> List[ABCDelegation]:
        result = []
        try:
            #self.lock.acquire()
            sid = str(slice_id) if slice_id is not None else None
            dlg_dict_list = self.db.get_delegations(slc_guid=sid, states=states)
            if self.lock.locked():
                self.lock.release()
            result = self._load_delegation_from_db(dlg_dict_list=dlg_dict_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def add_maintenance_properties(self, *, properties: dict):
        """
        Add maintenance properties
        @param properties properties
        """
        self.db.add_miscellaneous(name=self.MAINTENANCE, properties=properties)

    def update_maintenance_properties(self, *, properties: dict):
        """
        Update maintenance properties
        @param properties properties
        """
        self.db.update_miscellaneous(name=self.MAINTENANCE, properties=properties)

    def remove_maintenance_properties(self):
        """
        Remove maintenance properties
        """
        self.db.remove_miscellaneous(name=self.MAINTENANCE)

    def get_maintenance_properties(self) -> dict:
        """
        Get maintenance Properties
        @return properties
        """
        result = None
        try:
            result = self.db.get_miscellaneous(name=self.MAINTENANCE)
        except Exception as e:
            self.logger.error(e)
        return result

    def add_site(self, *, site: Site):
        self.logger.debug(f"Adding site {site.get_name()}")
        try:
            #self.lock.acquire()
            properties = pickle.dumps(site)
            self.db.add_site(site_name=site.get_name(), state=site.get_state().value, properties=properties)
            self.logger.debug(f"Site {site.get_name()} added")
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_site(self, *, site: Site):
        try:
            #self.lock.acquire()
            self.logger.debug(f"Updating site {site.get_name()}")
            properties = pickle.dumps(site)
            self.db.update_site(site_name=site.get_name(), state=site.get_state().value, properties=properties)
        finally:
            if self.lock.locked():
                self.lock.release()

    def remove_site(self, *, site_name: str):
        try:
            #self.lock.acquire()
            self.logger.debug(f"Removing site {site_name}")
            self.db.remove_site(site_name=site_name)
        finally:
            if self.lock.locked():
                self.lock.release()

    def _load_site_from_db(self, site_list: List[dict]) -> List[Site]:
        result = []
        if site_list is None:
            return result

        for s in site_list:
            pickled_site = s.get(Constants.PROPERTY_PICKLE_PROPERTIES)
            site = pickle.loads(pickled_site)
            result.append(site)

        return result

    def get_site(self, *, site_name: str) -> Site or None:
        try:
            #self.lock.acquire()
            site_list = self.db.get_site(site_name=site_name)
            if self.lock.locked():
                self.lock.release()
            if site_list is not None and len(site_list) > 0:
                return self._load_site_from_db(site_list=site_list)[0]
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        return None

    def get_sites(self) -> List[Site]:
        result = []
        try:
            #self.lock.acquire()
            site_list = self.db.get_sites()
            if self.lock.locked():
                self.lock.release()
            result = self._load_site_from_db(site_list=site_list)
        except Exception as e:
            self.logger.error(e)
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def add_config_mapping(self, key: str, config_mapping: ConfigurationMapping):
        try:
            #self.lock.acquire()
            properties = pickle.dumps(config_mapping)
            self.db.add_config_mapping(cfgm_type=key, act_id=self.actor_id, properties=properties)
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
            raise e
        finally:
            if self.lock.locked():
                self.lock.release()

    def get_config_mappings(self) -> List[ConfigurationMapping]:
        cfg_map_list = None
        result = []
        try:
            #self.lock.acquire()
            cfg_map_list = self.db.get_config_mappings(act_id=self.actor_id)
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        if cfg_map_list is not None:
            for c in cfg_map_list:
                pickled_cfg_map = c.get(Constants.PROPERTY_PICKLE_PROPERTIES)
                cfg_obj = pickle.loads(pickled_cfg_map)
                result.append(cfg_obj)
        return result

    def _load_poa_from_db(self, *, poa_dict_list: List[dict], include_res_info: bool) -> List[Poa]:
        result = []
        if poa_dict_list is None:
            return result

        for p in poa_dict_list:
            pickled_poa = p.get(Constants.PROPERTY_PICKLE_PROPERTIES)
            poa_obj = pickle.loads(pickled_poa)
            sliver_id = poa_obj.get_sliver_id()
            if include_res_info:
                reservations = self.get_reservations(rid=sliver_id)
                if reservations is not None:
                    poa_obj.restore(actor=self.actor, reservation=reservations[0])
            result.append(poa_obj)
        return result

    def get_poas(self, *, poa_id: str = None, email: str = None, sliver_id: ID = None, slice_id: ID = None,
                 project_id: str = None, limit: int = None, offset: int = None, last_update_time: datetime = None,
                 states: list[int] = None, include_res_info: bool = True) -> Union[List[Poa] or None]:
        result = []
        try:
            try:
                sliver_id_str = str(sliver_id) if sliver_id is not None else None
                slice_id_str = str(slice_id) if slice_id is not None else None

                poa_dict_list = self.db.get_poas(poa_guid=poa_id, email=email, sliver_id=sliver_id_str, limit=limit,
                                                 offset=offset, last_update_time=last_update_time,
                                                 project_id=project_id, slice_id=slice_id_str, states=states)
            finally:
                if self.lock.locked():
                    self.lock.release()
            if poa_dict_list is not None:
                result = self._load_poa_from_db(poa_dict_list=poa_dict_list, include_res_info=include_res_info)
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()
        return result

    def add_poa(self, *, poa: Poa):
        try:
            poa_list = self.get_poas(poa_id=poa.poa_id)
            if len(poa_list) > 0:
                raise DatabaseException(f"POA # {poa.get_poa_id()} already exists")

            properties = pickle.dumps(poa)
            email = None
            project_id = None
            slice_id = None
            if poa.get_slice().get_owner() is not None:
                email = poa.get_slice().get_owner().get_email()
                project_id = poa.get_slice().get_project_id()
                slice_id = str(poa.get_slice().get_slice_id())

            self.db.add_poa(poa_guid=poa.get_poa_id(), sliver_id=str(poa.get_sliver_id()), email=email,
                            project_id=project_id, slice_id=slice_id, properties=properties,
                            state=poa.get_state().value)
        finally:
            if self.lock.locked():
                self.lock.release()

    def update_poa(self, *, poa: Poa):
        try:
            #self.lock.acquire()
            properties = pickle.dumps(poa)
            self.db.update_poa(poa_guid=poa.poa_id, state=poa.get_state().value, properties=properties)
        finally:
            if self.lock.locked():
                self.lock.release()

    def remove_poa(self, *, poa_id: str):
        try:
            #self.lock.acquire()
            self.db.remove_poa(poa_guid=poa_id)
        finally:
            if self.lock.locked():
                self.lock.release()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin


class ABCContainerDatabase(ABC):
    """
    Interface for container-level databases. Defines the functions necessary for
    bootstrapping and recovering a container.
    """

    @abstractmethod
    def add_container_properties(self, *, properties: dict):
        """
        Adds the container properties  to the database.

        @param properties container properties dict.

        @throws Exception if an error occurs while accessing the database
        """

    @abstractmethod
    def get_container_properties(self) -> dict:
        """
        Retrieves the container properties.
        @return container properties
        @throws Exception if an error occurs while accessing the database
        """

    @abstractmethod
    def set_reset_state(self, *, value: bool):
        """
        Controls whether the database should reset its state.

        @param value TRUE if reset is required, FALSE otherwise
        """

    @abstractmethod
    def get_actors(self, *, name: str = None, actor_type: int = None) -> List[ABCActorMixin]:
        """
        Retrieves the actors defined in this container
        @return vector of properties
        @throws Exception in case of error
        """

    @abstractmethod
    def get_actor(self, *, actor_name: str) -> ABCActorMixin:
        """
        Retrieves the actors defined in this container
        @param actor_name actor name
        @return vector of properties
        @throws Exception in case of error
        """

    @abstractmethod
    def add_actor(self, *, actor: ABCActorMixin):
        """
        Adds a new actor record to the database
        @param actor actor to be added
        @throws Exception in case of error
        """

    @abstractmethod
    def remove_actor(self, *, actor_name: str):
        """
        Removes the specified actor record
        @param actor_name actor name
        @throws Exception in case of error
        """

    @abstractmethod
    def remove_actor_database(self, *, actor_name: str):
        """
        Destroy the database for this actor. Applies to actors storing their
        database on the same database server as the container database.
        @param actor_name actor name
        @throws Exception in case of error
        """

    @abstractmethod
    def add_time(self, *, properties: dict):
        """
        Adds the time record to the database
        @params properties: properties
        """

    @abstractmethod
    def get_time(self) -> dict:
        """
        Retrieves the time record from the database
        @return map of props
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import PolicyException
from fabric_cf.actor.core.util.resource_type import ResourceType

if TYPE_CHECKING:
    from fabric_cf.actor.core.policy.inventory_for_type import InventoryForType
    from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation


class Inventory:
    def __init__(self):
        self.map = {}

    def contains_type(self, *, resource_type: ResourceType):
        if resource_type is None:
            raise PolicyException(Constants.INVALID_ARGUMENT)

        if resource_type in self.map:
            return True

        return False

    def get(self, *, resource_type: ResourceType) -> InventoryForType:
        if resource_type is None:
            raise PolicyException(Constants.INVALID_ARGUMENT)

        return self.map.get(resource_type, None)

    def remove(self, *, source: ABCClientReservation):
        """
        Removes the inventory derived from the specified source.
        @param source source reservation
        @return true if the inventory was update, false otherwise
        """
        rtype = source.get_type()

        if rtype in self.map:
            inv = self.map[rtype]
            if inv.source == source:
                self.map.pop(rtype)
                return True

        return False

    def get_inventory(self) -> dict:
        return self.map

    def remove_inventory_by_type(self, *, rtype: ResourceType):
        if rtype in self.map:
            self.map.pop(rtype)

    def add_inventory_by_type(self, *, inventory: InventoryForType, rtype: ResourceType):
        if rtype not in self.map:
            self.map[rtype] = inventory



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING, List
import logging
import queue
import threading
import time
import traceback


from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro
from fabric_cf.actor.core.util.iterable_queue import IterableQueue

if TYPE_CHECKING:
    from fabric_cf.actor.core.proxies.kafka.services.actor_service import ActorService
    from fabric_cf.actor.core.manage.kafka.services.kafka_actor_service import KafkaActorService


class RPCConsumerException(Exception):
    pass


class RPCConsumer:
    MANAGEMENT_MESSAGES = [AbcMessageAvro.claim_resources,
                           AbcMessageAvro.reclaim_resources,
                           AbcMessageAvro.get_slices_request,
                           AbcMessageAvro.get_reservations_request,
                           AbcMessageAvro.get_reservations_state_request,
                           AbcMessageAvro.get_delegations,
                           AbcMessageAvro.get_reservation_units_request,
                           AbcMessageAvro.get_unit_request,
                           AbcMessageAvro.get_broker_query_model_request,
                           AbcMessageAvro.add_slice,
                           AbcMessageAvro.update_slice,
                           AbcMessageAvro.remove_slice,
                           AbcMessageAvro.close_reservations,
                           AbcMessageAvro.update_reservation,
                           AbcMessageAvro.remove_reservation,
                           AbcMessageAvro.extend_reservation,
                           AbcMessageAvro.close_delegations,
                           AbcMessageAvro.remove_delegation,
                           AbcMessageAvro.maintenance_request,
                           AbcMessageAvro.get_sites_request]

    def __init__(self, *, kafka_service: ActorService, kafka_mgmt_service: KafkaActorService,
                 logger: logging.Logger = None):
        self.message_queue = queue.Queue()
        self.thread_lock = threading.Lock()
        self.condition = threading.Condition()
        self.thread = None
        self.shutdown = False
        self.name = self.__class__.__name__
        if logger is None:
            self.logger = logging.Logger(self.name)
        else:
            self.logger = logger
        self.kafka_service = kafka_service
        self.kafka_mgmt_service = kafka_mgmt_service

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['message_queue']
        del state['thread_lock']
        del state['condition']
        del state['thread']
        del state['shutdown']
        del state['logger']

    def __setstate__(self, state):
        self.message_queue = queue.Queue()
        self.thread_lock = threading.Lock()
        self.condition = threading.Condition()
        self.thread = None
        self.shutdown = False
        self.logger = None

    def set_logger(self, *, logger: logging.Logger):
        self.logger = logger

    def start(self):
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise RPCConsumerException(f"{self.name} has already been started")

            self.thread = threading.Thread(target=self.__run, name=self.name, daemon=True)
            self.thread.start()
            self.logger.debug(f"{self.name} has been started")
        finally:
            self.thread_lock.release()

    def stop(self):
        self.shutdown = True
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning(f"It seems that the f{self.name} thread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error(f"Could not join {self.name} thread {e}")
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def enqueue(self, incoming):
        try:
            self.message_queue.put_nowait(incoming)
            with self.condition:
                self.condition.notify_all()
            self.logger.debug("Added message to queue {}".format(incoming.__class__.__name__))
        except Exception as e:
            self.logger.error(f"Failed to queue message: {incoming.__class__.__name__} e: {e}")

    def __dequeue(self, queue_obj: queue.Queue):
        events = []
        if not queue_obj.empty():
            try:
                for event in IterableQueue(source_queue=queue_obj):
                    events.append(event)
            except Exception as e:
                self.logger.error(f"Error while adding message to queue! e: {e}")
                self.logger.error(traceback.format_exc())
        return events

    def __process_messages(self, *, messages: list):
        for message in messages:
            try:
                begin = time.time()
                if message.get_message_name() in self.MANAGEMENT_MESSAGES:
                    self.kafka_mgmt_service.process(message=message)
                else:
                    self.kafka_service.process(message=message)
                diff = int(time.time() - begin)
                if diff > 0:
                    self.logger.info(f"Event {message.__class__.__name__} TIME: {diff}")
            except Exception as e:
                self.logger.error(f"Error while processing message {type(message)}, {e}")
                self.logger.error(traceback.format_exc())

    def __run(self):
        while True:
            with self.condition:
                while not self.shutdown and self.message_queue.empty():
                    self.condition.wait()

            if self.shutdown:
                break

            if self.shutdown:
                self.logger.info(f"{self.name} exiting")
                return

            messages = self.__dequeue(self.message_queue)
            self.__process_messages(messages=messages)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server.models.poa_post_data import PoaPostData  # noqa: F401,E501
from fabric_cf.orchestrator.swagger_server import util


class PoaPost(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, operation: str=None, data: PoaPostData=None):  # noqa: E501
        """PoaPost - a model defined in Swagger

        :param operation: The operation of this PoaPost.  # noqa: E501
        :type operation: str
        :param data: The data of this PoaPost.  # noqa: E501
        :type data: PoaPostData
        """
        self.swagger_types = {
            'operation': str,
            'data': PoaPostData
        }

        self.attribute_map = {
            'operation': 'operation',
            'data': 'data'
        }
        self._operation = operation
        self._data = data

    @classmethod
    def from_dict(cls, dikt) -> 'PoaPost':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The poa_post of this PoaPost.  # noqa: E501
        :rtype: PoaPost
        """
        return util.deserialize_model(dikt, cls)

    @property
    def operation(self) -> str:
        """Gets the operation of this PoaPost.


        :return: The operation of this PoaPost.
        :rtype: str
        """
        return self._operation

    @operation.setter
    def operation(self, operation: str):
        """Sets the operation of this PoaPost.


        :param operation: The operation of this PoaPost.
        :type operation: str
        """
        allowed_values = ["cpuinfo", "numainfo", "cpupin", "numatune", "reboot", "addkey", "removekey", "rescan"]  # noqa: E501
        if operation not in allowed_values:
            raise ValueError(
                "Invalid value for `operation` ({0}), must be one of {1}"
                .format(operation, allowed_values)
            )

        self._operation = operation

    @property
    def data(self) -> PoaPostData:
        """Gets the data of this PoaPost.


        :return: The data of this PoaPost.
        :rtype: PoaPostData
        """
        return self._data

    @data.setter
    def data(self, data: PoaPostData):
        """Sets the data of this PoaPost.


        :param data: The data of this PoaPost.
        :type data: PoaPostData
        """

        self._data = data



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import sched
import sys
import threading
import traceback
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import logging
import os

from fabric_cf.actor.core.util.quota_mgr import QuotaMgr
from fim.graph.neo4j_property_graph import Neo4jGraphImporter
from fim.graph.resources.abc_arm import ABCARMPropertyGraph
from fss_utils.jwt_validate import JWTValidator

from fabric_cf.actor.core.common.event_logger import EventLoggerSingleton
from fabric_cf.actor.core.common.exceptions import InitializationException
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.container.container import Container
from fabric_cf.actor.core.util.log_helper import LogHelper
from fabric_cf.actor.security.token_validator import TokenValidator

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_container import ABCActorContainer
    from fabric_cf.actor.boot.configuration import Configuration

logging.TRACE = 5
logging.addLevelName(logging.TRACE, "TRACE")
logging.Logger.trace = lambda inst, msg, *args, **kwargs: inst.log(logging.TRACE, msg, *args, **kwargs)
logging.trace = lambda msg, *args, **kwargs: logging.log(logging.TRACE, msg, *args, **kwargs)


class Globals:
    config_file = Constants.CONFIGURATION_FILE
    RPC_TIMEOUT = 0

    def __init__(self):
        self.config = None
        self.log = None
        self.initialized = False
        self.started = False
        self.start_completed = False
        self.container = None
        self.properties = None
        self.timer_scheduler = sched.scheduler()
        self.timer_thread = None
        self.timer_condition = threading.Condition()
        self.lock = threading.Lock()
        self.jwt_validator = None
        self.token_validator = None
        self.quota_mgr = None

    def make_logger(self):
        """
        Detects the path and level for the log file from the actor config and sets
        up a logger. Instead of detecting the path and/or level from the
        config, a custom path and/or level for the log file can be passed as
        optional arguments.

       :return: logging.Logger object
        """
        log_config = self.config.get_global_config().get_logging()
        if log_config is None:
            raise RuntimeError('No logging  config information available')

        EventLoggerSingleton.get().make_logger(log_config=log_config)

        log_dir = log_config.get(Constants.PROPERTY_CONF_LOG_DIRECTORY, ".")
        log_file = log_config.get(Constants.PROPERTY_CONF_LOG_FILE, "actor.log")
        log_level = log_config.get(Constants.PROPERTY_CONF_LOG_LEVEL, None)
        log_retain = int(log_config.get(Constants.PROPERTY_CONF_LOG_RETAIN, 50))
        log_size = int(log_config.get(Constants.PROPERTY_CONF_LOG_SIZE, 5000000))
        logger = log_config.get(Constants.PROPERTY_CONF_LOGGER, "actor")

        return LogHelper.make_logger(log_dir=log_dir, log_file=log_file, log_level=log_level, log_retain=log_retain,
                                     log_size=log_size, logger=logger)

    @staticmethod
    def delete_super_block():
        """
        Delete Super block file
        """
        if os.path.isfile(Constants.SUPERBLOCK_LOCATION):
            os.remove(Constants.SUPERBLOCK_LOCATION)

    @staticmethod
    def can_reload_model() -> bool:
        if os.path.isfile(Constants.MODEL_RELOAD_LOCATION):
            return True
        return False

    @staticmethod
    def delete_reload_model_state_file():
        """
        Delete reload model state file
        """
        if os.path.isfile(Constants.MODEL_RELOAD_LOCATION):
            os.remove(Constants.MODEL_RELOAD_LOCATION)

    def cleanup_neo4j(self):
        """
        Cleanup Neo4j on clean restart
        """
        self.log.debug("Cleanup Neo4j database started")
        config = self.get_config().get_neo4j_config()
        neo4j_graph_importer = Neo4jGraphImporter(url=config["url"], user=config["user"],
                                                  pswd=config["pass"],
                                                  import_host_dir=config["import_host_dir"],
                                                  import_dir=config["import_dir"])
        neo4j_graph_importer.delete_all_graphs()
        self.log.debug("Cleanup Neo4j database completed")

    def check_and_reload_model(self, *, graph_id) -> ABCARMPropertyGraph or None:
        """
            Reload Neo4j on model restart
        """
        if not self.can_reload_model():
            return None
        self.cleanup_neo4j()
        self.log.debug(f"Reload Neo4j database started {graph_id}")
        from fabric_cf.actor.fim.fim_helper import FimHelper
        arm_graph = FimHelper.get_arm_graph_from_file(filename=self.get_config().get_actor_config().get_substrate_file(),
                                                      graph_id=graph_id)
        self.log.debug(f"Reload Neo4j database completed {graph_id}")
        return arm_graph

    def fail(self, *, e: Exception):
        """
        Fail the Actor
        @param e exception
        """
        self.log.error("Critical error: Actor failed to initialize {}".format(e))
        sys.exit(-1)

    def initialize(self):
        """
        Initialize the container and actor
        """
        try:
            self.lock.acquire()
            if not self.initialized:
                self.load_config()
                self.log = self.make_logger()
                self.log.info("Checking if connection to Kafka broker can be established")
                admin_kafka_client = self.get_kafka_admin_client()
                admin_kafka_client.list_topics()
                self.log.info("Connection to Kafka broker established successfully")
                self.load_validators()
                self.log.info("Main initialization complete.")
                self.initialized = True
        finally:
            self.lock.release()

    def load_validators(self):
        oauth_config = self.config.get_oauth_config()
        CREDMGR_CERTS = oauth_config.get(Constants.PROPERTY_CONF_O_AUTH_JWKS_URL, None)
        CREDMGR_KEY_REFRESH = oauth_config.get(Constants.PROPERTY_CONF_O_AUTH_KEY_REFRESH, None)
        CREDMGR_TRL_REFRESH = oauth_config.get(Constants.PROPERTY_CONF_O_AUTH_TRL_REFRESH, '00:01:00')
        self.log.info(f'Initializing JWT Validator to use {CREDMGR_CERTS} endpoint, '
                      f'refreshing keys every {CREDMGR_KEY_REFRESH} HH:MM:SS refreshing '
                      f'token revoke list every {CREDMGR_TRL_REFRESH} HH:MM:SS')
        t = datetime.strptime(CREDMGR_KEY_REFRESH, "%H:%M:%S")
        self.jwt_validator = JWTValidator(url=CREDMGR_CERTS,
                                          refresh_period=timedelta(hours=t.hour, minutes=t.minute, seconds=t.second))
        from urllib.parse import urlparse
        t = datetime.strptime(CREDMGR_KEY_REFRESH, "%H:%M:%S")
        self.token_validator = TokenValidator(credmgr_host=str(urlparse(CREDMGR_CERTS).hostname),
                                              refresh_period=timedelta(hours=t.hour, minutes=t.minute, seconds=t.second),
                                              jwt_validator=self.jwt_validator)

        core_api = self.config.get_core_api_config()
        if core_api.get("enable", False):
            self.quota_mgr = QuotaMgr(core_api_host=core_api.get(Constants.PROPERTY_CONF_HOST),
                                      token=core_api.get(Constants.TOKEN, ""),
                                      logger=self.log)

    def load_config(self):
        """
        Load the configuration
        """
        try:
            from fabric_cf.actor.boot.configuration_loader import ConfigurationLoader
            loader = ConfigurationLoader(path=self.config_file)
            self.config = loader.read_configuration()
            self.RPC_TIMEOUT = self.config.get_rpc_request_timeout_seconds()
        except Exception as e:
            raise RuntimeError("Unable to parse configuration file {}".format(e))

    def get_jwt_validator(self) -> JWTValidator:
        return self.jwt_validator

    def get_token_validator(self) -> TokenValidator:
        return self.token_validator

    def get_quota_mgr(self) -> QuotaMgr:
        return self.quota_mgr

    def get_container(self) -> ABCActorContainer:
        """
        Get the container
        @return container
        """
        if not self.initialized:
            raise InitializationException(Constants.UNINITIALIZED_STATE)
        return self.container

    def get_config(self) -> Configuration:
        """
        Get the configuration
        @return config
        """
        if not self.initialized:
            raise InitializationException(Constants.UNINITIALIZED_STATE)
        return self.config

    def get_log_config(self) -> dict:
        """
        Get the Log configuration
        @return dict
        """
        return self.get_config().get_log_config()

    def get_kafka_config_admin_client(self) -> dict:
        """
        Get Kafka Config Admin Client
        @retun admin client config
        """
        if self.config is None or self.config.get_runtime_config() is None:
            return None

        sasl_username = self.config.get_kafka_prod_user_name()
        sasl_password = self.config.get_kafka_prod_user_pwd()
        sasl_mechanism = self.config.get_kafka_sasl_mechanism()

        conf = {Constants.BOOTSTRAP_SERVERS: self.config.get_kafka_server(),
                Constants.SECURITY_PROTOCOL: self.config.get_kafka_security_protocol(),
                Constants.SSL_CA_LOCATION: self.config.get_kafka_ssl_ca_location(),
                Constants.SSL_CERTIFICATE_LOCATION: self.config.get_kafka_ssl_cert_location(),
                Constants.SSL_KEY_LOCATION: self.config.get_kafka_ssl_key_location(),
                Constants.SSL_KEY_PASSWORD: self.config.get_kafka_ssl_key_password()}

        if sasl_username is not None and sasl_username != '' and sasl_password is not None and sasl_password != '':
            conf[Constants.SASL_USERNAME] = sasl_username
            conf[Constants.SASL_PASSWORD] = sasl_password
            conf[Constants.SASL_MECHANISM] = sasl_mechanism

        return conf

    def get_kafka_config_schema_registry_client(self) -> dict:
        """
        Get Producer Config
        @return producer config
        """
        if self.config and self.config.get_runtime_config():
            return {"url": self.config.get_kafka_schema_registry() }

    def get_kafka_config_producer(self) -> dict:
        """
        Get Producer Config
        @return producer config
        """
        if self.config is None or self.config.get_runtime_config() is None:
            return None

        sasl_username = self.config.get_kafka_prod_user_name()
        sasl_password = self.config.get_kafka_prod_user_pwd()
        sasl_mechanism = self.config.get_kafka_sasl_mechanism()

        conf = {Constants.BOOTSTRAP_SERVERS: self.config.get_kafka_server(),
                Constants.SECURITY_PROTOCOL: self.config.get_kafka_security_protocol(),
                Constants.SSL_CA_LOCATION: self.config.get_kafka_ssl_ca_location(),
                Constants.SSL_CERTIFICATE_LOCATION: self.config.get_kafka_ssl_cert_location(),
                Constants.SSL_KEY_LOCATION: self.config.get_kafka_ssl_key_location(),
                Constants.SSL_KEY_PASSWORD: self.config.get_kafka_ssl_key_password(),
                Constants.SCHEMA_REGISTRY_URL: self.config.get_kafka_schema_registry(),
                Constants.PROPERTY_CONF_KAFKA_REQUEST_TIMEOUT_MS: self.config.get_kafka_request_timeout_ms(),
                Constants.PROPERTY_CONF_KAFKA_MAX_MESSAGE_SIZE: self.config.get_kafka_max_message_size()}

        if sasl_username is not None and sasl_username != '' and sasl_password is not None and sasl_password != '':
            conf[Constants.SASL_USERNAME] = sasl_username
            conf[Constants.SASL_PASSWORD] = sasl_password
            conf[Constants.SASL_MECHANISM] = sasl_mechanism

        return conf

    def get_kafka_consumer_poll_timeout(self):
        if self.get_config():
            return self.get_config().get_kafka_consumer_poll_timeout()

    def get_kafka_consumer_commit_batch_size(self):
        if self.get_config():
            return self.get_config().get_kafka_consumer_commit_batch_size()

    def get_kafka_consumer_auto_commit_interval(self):
        if self.get_config():
            return self.get_config().get_kafka_consumer_auto_commit_interval()

    def get_kafka_consumer_enable_auto_commit(self):
        if self.get_config():
            return self.get_config().get_kafka_consumer_enable_auto_commit()

    def get_kafka_key_schema_location(self):
        if self.get_config():
            return self.get_config().get_kafka_key_schema_location()

    def get_kafka_value_schema_location(self):
        if self.get_config():
            return self.get_config().get_kafka_value_schema_location()

    def get_kafka_config_consumer(self) -> dict:
        """
        Get Consumer config
        @return consumer config
        """
        if self.config and self.config.get_runtime_config():
            conf = self.get_kafka_config_producer()

            group_id = self.get_config().get_kafka_cons_group_id()

            conf['auto.offset.reset'] = 'earliest'
            enable_auto_commit = self.get_kafka_consumer_enable_auto_commit()
            if enable_auto_commit:
                conf[Constants.PROPERTY_CONF_KAFKA_AUTO_COMMIT_INTERVAL] = self.get_kafka_consumer_auto_commit_interval()
            else:
                conf[Constants.PROPERTY_CONF_KAFKA_ENABLE_AUTO_COMMIT] = enable_auto_commit

            sasl_username = self.get_config().get_kafka_cons_user_name()
            sasl_password = self.get_config().get_kafka_cons_user_pwd()

            if sasl_username is not None and sasl_username != '' and sasl_password is not None and sasl_password != '':
                conf[Constants.SASL_USERNAME] = sasl_username
                conf[Constants.SASL_PASSWORD] = sasl_password
            conf[Constants.GROUP_ID] = group_id
            conf[Constants.PROPERTY_CONF_KAFKA_FETCH_MAX_MESSAGE_SIZE] = self.get_config().get_kafka_max_message_size()
            return conf

    def get_kafka_producer(self):
        """
        Create and return a kafka producer
        @return producer
        """
        conf = self.get_kafka_config_producer()
        key_schema_file = self.config.get_kafka_key_schema_location()
        value_schema_file = self.config.get_kafka_value_schema_location()

        from fabric_mb.message_bus.producer import AvroProducerApi
        producer = AvroProducerApi(producer_conf=conf,
                                   key_schema_location=key_schema_file,
                                   #schema_registry_conf=self.get_kafka_config_schema_registry_client(),
                                   value_schema_location=value_schema_file, logger=self.get_logger())
        return producer

    def get_kafka_producer_with_poller(self, *, actor):
        """
        Create and return a kafka producer
        @return producer
        """
        conf = self.get_kafka_config_producer()
        key_schema_file = self.config.get_kafka_key_schema_location()
        value_schema_file = self.config.get_kafka_value_schema_location()

        from fabric_cf.actor.core.container.rpc_producer import RPCProducer
        producer = RPCProducer(producer_conf=conf, key_schema_location=key_schema_file,
                               value_schema_location=value_schema_file, logger=self.get_logger(),
                               actor=actor, retries=self.config.get_rpc_retries())
        # Uncomment for 1.7
        #producer = RPCProducer(producer_conf=conf, schema_registry_conf=self.get_kafka_config_schema_registry_client(),
        #                       value_schema_location=value_schema_file, logger=self.get_logger(),
        #                       actor=actor, retries=self.config.get_rpc_retries())
        return producer

    def get_simple_kafka_producer(self):
        """
        Create and return a kafka producer
        @return producer
        """
        conf = self.get_kafka_config_producer()
        if conf is not None:
            conf.pop(Constants.SCHEMA_REGISTRY_URL)

        bqm_config = self.config.get_global_config().get_bqm_config()
        if bqm_config is not None:
            prod_user_name = bqm_config.get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME, None)
            prod_password = bqm_config.get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD, None)
            if prod_user_name is not None and prod_password is not None:
                conf[Constants.SASL_USERNAME] = prod_user_name
                conf[Constants.SASL_PASSWORD] = prod_password

        from confluent_kafka import Producer
        producer = Producer(conf)
        return producer

    def get_kafka_admin_client(self):
        """
        Create and return a kafka admin client
        @return admin client
        """
        from fabric_mb.message_bus.admin import AdminApi
        conf = self.get_kafka_config_admin_client()
        admin = AdminApi(conf=conf)
        return admin

    def get_logger(self):
        """
        Get logger
        @return logger
        """
        if not self.initialized:
            raise InitializationException(Constants.UNINITIALIZED_STATE)

        if self.log is None:
            self.log = self.make_logger()
        return self.log

    def start(self, *, force_fresh: bool):
        """
        Start CF Actor
        @param force_fresh true if clean restart, false if stateful restart
        """
        try:
            try:
                self.lock.acquire()
                if self.started:
                    return
                self.started = True
                self.start_completed = False
                if force_fresh:
                    self.delete_super_block()
            finally:
                self.lock.release()

            self.initialize()
            self.start_timer_thread()
            try:
                self.lock.acquire()
                self.container = Container()
                self.log.info("Successfully instantiated the container implementation.")
                self.log.info("Initializing container")
                self.container.initialize(config=self.config)
                self.log.info("Successfully initialized the container")
                self.start_completed = True
            finally:
                self.lock.release()
        except Exception as e:
            self.log.error(traceback.format_exc())
            self.fail(e=e)

    def stop(self):
        """
        Stop the Actor
        """
        try:
            self.lock.acquire()
            if not self.started:
                return
            self.log.info("Stopping Actor")
            self.started = False
            self.stop_timer_thread()
            self.get_container().shutdown()
        except Exception as e:
            self.log.error("Error while shutting down: {}".format(e))
        finally:
            self.lock.release()

    def start_timer_thread(self):
        """
        Start the timer thread
        """
        if self.timer_thread is not None:
            raise RuntimeError("This timer thread has already been started")

        self.timer_thread = threading.Thread(target=self.timer_loop)
        self.timer_thread.setName('GlobalTimer')
        self.timer_thread.setDaemon(True)
        self.timer_thread.start()

    def stop_timer_thread(self):
        """
        Stop timer thread
        """
        temp = self.timer_thread
        self.timer_thread = None
        if temp is not None:
            self.log.warning("It seems that the timer thread is running. Interrupting it")
            try:
                with self.timer_condition:
                    self.timer_condition.notify_all()
                temp.join()
            except Exception as e:
                self.log.error("Could not join timer thread {}".format(e))

    def timer_loop(self):
        """
        Timer thread run function
        """
        self.log.debug(f"Timer thread started")
        while True:
            with self.timer_condition:
                while self.timer_scheduler.empty() and self.started:
                    try:
                        #self.log.debug("Waiting for condition")
                        self.timer_condition.wait()
                    except InterruptedError as e:
                        self.log.error(traceback.format_exc())
                        self.log.error(f" Timer thread interrupted. Exiting {e}")
                        return

                    if not self.started:
                        self.log.info(f"Timer thread exiting")
                        return

                    self.timer_condition.notify_all()

                if not self.timer_scheduler.empty():
                    #self.log.debug(f"Executing Scheduled items")
                    self.timer_scheduler.run(blocking=False)


class GlobalsSingleton:
    """
    Global Singleton class
    """
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise InitializationException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = Globals()
        return self.__instance

    get = classmethod(get)



#!/usr/bin/env python3
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import argparse
import logging
import re
import tempfile
import traceback
import os
from datetime import datetime, timezone
from logging.handlers import RotatingFileHandler

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver

from fabric_cf.actor.core.kernel.slice import SliceTypes
from fabric_cf.actor.core.plugins.db.actor_database import ActorDatabase
from fabric_cf.actor.core.container.globals import Globals, GlobalsSingleton
from fabric_cf.actor.core.policy.inventory_for_type import InventoryForType
from fabric_reports_client.reports_api import ReportsApi


LAST_EXPORT_FILE = "./last_export_time.txt"


class ExportScript:
    """
    CLI interface to fetch data from Postgres and push to reports db via reportsApi.
    """

    def __init__(self, config_file: str, batch_size=1000):
        self.logger = logging.getLogger("export")
        file_handler = RotatingFileHandler('/var/log/actor/export.log', backupCount=5, maxBytes=50000)
        logging.basicConfig(level=logging.INFO,
                            format="%(asctime)s [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s",
                            handlers=[logging.StreamHandler(), file_handler])

        Globals.config_file = config_file
        GlobalsSingleton.get().load_config()
        GlobalsSingleton.get().initialized = True
        self.config = GlobalsSingleton.get().get_config()

        db_conf = self.config.get_global_config().get_database()
        self.src_db = ActorDatabase(user=db_conf.get("db-user"),
                                    password=db_conf.get("db-password"),
                                    database=db_conf.get("db-name"),
                                    db_host=db_conf.get("db-host"),
                                    logger=self.logger)

        self.reports_conf = self.config.get_reports_api()
        self.temp_token_file = self._create_temp_token_file(self.reports_conf.get("token"))
        self.reports_api = ReportsApi(base_url=self.reports_conf.get("host"), token_file=self.temp_token_file)

        self.actor_config = self.config.get_actor_config()
        self.batch_size = batch_size
        self.last_export_time = self.get_last_export_time()

    def _create_temp_token_file(self, token: str) -> str:
        fd, path = tempfile.mkstemp(prefix="token_", suffix=".json")
        with os.fdopen(fd, 'w') as f:
            f.write(f'{{"id_token": "{token}"}}')
        return path

    def get_last_export_time(self):
        """
        Reads the last export timestamp from a file.
        """
        if os.path.exists(LAST_EXPORT_FILE):
            with open(LAST_EXPORT_FILE, "r") as f:
                timestamp_str = f.read().strip()
                try:
                    return datetime.fromisoformat(timestamp_str)
                except ValueError:
                    self.logger.warning("Invalid timestamp in last export file, defaulting to epoch.")
        return datetime(1970, 1, 1, tzinfo=timezone.utc)  # Default to epoch

    def update_last_export_time(self, new_timestamp):
        """
        Updates the last export timestamp file with the current run time.
        """
        with open(LAST_EXPORT_FILE, "w") as f:
            f.write(new_timestamp.isoformat())

    def export(self):
        """
        Exports only the slices updated after the last execution timestamp.
        """
        try:
            if not self.reports_conf.get("enable", False):
                self.logger.info(f"Exiting export process... not enabled {self.reports_conf}")
                return
            actor_type = self.actor_config.get_type()
            if actor_type.lower() != ActorType.Orchestrator.name.lower():
                self.logger.info(f"Exiting export process... not orchestrator")
                return

            self.logger.info(f"Starting export process... Last export was at {self.last_export_time}")

            offset = 0
            new_timestamp = datetime.now(timezone.utc)

            while True:
                self.logger.info(f"Fetching slices from offset {offset} (batch size: {self.batch_size})")
                slices = self.src_db.get_slices(offset=offset, limit=self.batch_size, slc_type=[SliceTypes.ClientSlice],
                                                updated_after=self.last_export_time)  # Fetch only updated slices

                if not slices or len(slices) == 0:
                    self.logger.info("No more slices to process. Export complete.")
                    break  # Stop when no more slices are found

                for slice_object in slices:
                    try:
                        slice_guid = str(slice_object.get_slice_id())
                        self.reports_api.post_slice(slice_id=slice_guid,
                                                    slice_payload={
                                                        "project_id": slice_object.get_project_id(),
                                                        "project_name": slice_object.get_project_name(),
                                                        "user_id": slice_object.get_owner().get_oidc_sub_claim(),
                                                        "user_email": slice_object.get_owner().get_email(),
                                                        "slice_id": slice_guid,
                                                        "slice_name": slice_object.get_name(),
                                                        "state": slice_object.get_state().name,
                                                        "lease_start": slice_object.get_lease_start().isoformat(),
                                                        "lease_end": slice_object.get_lease_end().isoformat()

                                                    })

                        for reservation in self.src_db.get_reservations(slice_id=slice_object.get_slice_id()):
                            error_message = reservation.get_error_message()
                            sliver_guid = str(reservation.get_reservation_id())
                            sliver = InventoryForType.get_allocated_sliver(reservation=reservation)
                            site_name = None
                            host_name = None
                            ip_subnet = None
                            core = None
                            ram = None
                            disk = None
                            image = None
                            bw = None
                            node_id = None

                            if isinstance(sliver, NodeSliver):
                                site_name = sliver.get_site()
                                if sliver.label_allocations and sliver.label_allocations.instance_parent:
                                    host_name = sliver.label_allocations.instance_parent
                                ip_subnet = str(sliver.management_ip)
                                image = sliver.image_ref
                                node_id = str(reservation.get_graph_node_id())

                                if sliver.capacity_allocations:
                                    core = sliver.capacity_allocations.core
                                    ram = sliver.capacity_allocations.ram
                                    disk = sliver.capacity_allocations.disk

                            elif isinstance(sliver, NetworkServiceSliver):
                                site_name = sliver.get_site()
                                if sliver.get_gateway():
                                    ip_subnet = str(sliver.get_gateway().subnet)
                                if sliver.capacities:
                                    bw = sliver.capacities.bw

                            sliver_payload = {
                                "project_id": slice_object.get_project_id(),
                                "project_name": slice_object.get_project_name(),
                                "slice_id": slice_guid,
                                "slice_name": slice_object.get_name(),
                                "user_id": slice_object.get_owner().get_oidc_sub_claim(),
                                "user_email": slice_object.get_owner().get_email(),
                                "host": host_name,
                                "site": site_name,
                                "sliver_id": sliver_guid,
                                "node_id": node_id,
                                "state": reservation.get_state().name.lower(),
                                "sliver_type": str(reservation.get_type()).lower(),
                                "ip_subnet": ip_subnet,
                                "error": error_message,
                                "image": image,
                                "core": core,
                                "ram": ram,
                                "disk": disk,
                                "bandwidth": bw,
                                "lease_start": reservation.get_term().get_start_time().isoformat(),
                                "lease_end": reservation.get_term().get_end_time().isoformat(),
                                "interfaces": {
                                    "data": []
                                },
                                "components": {
                                    "data": []
                                }
                            }
                            if isinstance(sliver, NodeSliver) and sliver.attached_components_info:
                                components = []
                                for component in sliver.attached_components_info.devices.values():
                                    bdfs = component.labels.bdf if component.labels and component.labels.bdf else None
                                    if bdfs and not isinstance(bdfs, list):
                                        bdfs = [bdfs]
                                    node_id = None
                                    component_node_id = None

                                    sliver_map = sliver.get_node_map()
                                    if sliver_map:
                                        _, node_id = sliver_map

                                    component_map = component.get_node_map()
                                    if component_map:
                                        _, component_node_id = component_map

                                    components.append({
                                        "component_id": component.node_id,
                                        "node_id": node_id,
                                        "component_node_id": component_node_id,
                                        "type": str(component.get_type()).lower(),
                                        "model": str(component.get_model()).lower(),
                                        "bdfs": bdfs
                                    })
                                if len(components):
                                    sliver_payload["components"]["data"] = components

                            if isinstance(sliver, NetworkServiceSliver) and sliver.interface_info:
                                interfaces = []
                                for ifs in sliver.interface_info.interfaces.values():
                                    site = None
                                    vlan = ifs.labels.vlan if ifs.labels else None
                                    if not vlan and ifs.label_allocations:
                                        vlan = ifs.label_allocations.vlan

                                    bdf = ifs.labels.bdf if ifs.labels else None
                                    if not bdf and ifs.label_allocations:
                                        bdf = ifs.label_allocations.bdf

                                    local_name = ifs.labels.local_name if ifs.labels else None
                                    if not local_name and ifs.label_allocations:
                                        local_name = ifs.label_allocations.local_name

                                    device_name = ifs.labels.device_name if ifs.labels else None
                                    if not device_name and ifs.label_allocations:
                                        device_name = ifs.label_allocations.device_name
                                        if device_name:
                                            result = re.findall(r'\b([\w]+)-data-sw\b', device_name)
                                            if result and len(result) > 0:
                                                site = result[0]

                                    interfaces.append({
                                        "interface_id": ifs.node_id,
                                        "site": site,
                                        "vlan": vlan,
                                        "bdf": bdf,
                                        "local_name": local_name,
                                        "device_name": device_name,
                                        "name": ifs.get_name()
                                    })
                                if len(interfaces):
                                    sliver_payload["interfaces"]["data"] = interfaces

                            self.reports_api.post_sliver(slice_id=slice_guid,
                                                         sliver_id=sliver_guid,
                                                         sliver_payload=sliver_payload)

                    except Exception as slice_error:
                        self.logger.error(f"Error processing slice {slice_object.get_slice_id()}: {slice_error}")
                        traceback.print_exc()
                offset += self.batch_size  # Move to the next batch

            self.logger.info(f"Updating last export time to {new_timestamp}")
            self.update_last_export_time(new_timestamp)
            self.logger.info("Export process completed successfully!")

        except Exception as e:
            self.logger.error(f"Exception occurred during export: {e}")
            traceback.print_exc()
        finally:
            if os.path.exists(self.temp_token_file):
                os.remove(self.temp_token_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Export data from Postgres to SQLAlchemy DB via DatabaseManager")
    parser.add_argument("--config_file", default="/etc/fabric/actor/config/config.yaml", help="Path to config file")
    parser.add_argument("--batch_size", type=int, default=1000, help="Number of slices to process per batch")

    args = parser.parse_args()
    exporter = ExportScript(config_file=args.config_file, batch_size=args.batch_size)
    exporter.export()



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server.models.poa_post_data_keys import PoaPostDataKeys  # noqa: F401,E501
from fabric_cf.orchestrator.swagger_server.models.poa_post_data_vcpu_cpu_map import PoaPostDataVcpuCpuMap  # noqa: F401,E501
from fabric_cf.orchestrator.swagger_server import util


class PoaPostData(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, vcpu_cpu_map: List[PoaPostDataVcpuCpuMap]=None, node_set: List[str]=None, bdf: List[str]=None, keys: List[PoaPostDataKeys]=None):  # noqa: E501
        """PoaPostData - a model defined in Swagger

        :param vcpu_cpu_map: The vcpu_cpu_map of this PoaPostData.  # noqa: E501
        :type vcpu_cpu_map: List[PoaPostDataVcpuCpuMap]
        :param node_set: The node_set of this PoaPostData.  # noqa: E501
        :type node_set: List[str]
        :param bdf: The bdf of this PoaPostData.  # noqa: E501
        :type bdf: List[str]
        :param keys: The keys of this PoaPostData.  # noqa: E501
        :type keys: List[PoaPostDataKeys]
        """
        self.swagger_types = {
            'vcpu_cpu_map': List[PoaPostDataVcpuCpuMap],
            'node_set': List[str],
            'bdf': List[str],
            'keys': List[PoaPostDataKeys]
        }

        self.attribute_map = {
            'vcpu_cpu_map': 'vcpu_cpu_map',
            'node_set': 'node_set',
            'bdf': 'bdf',
            'keys': 'keys'
        }
        self._vcpu_cpu_map = vcpu_cpu_map
        self._node_set = node_set
        self._bdf = bdf
        self._keys = keys

    @classmethod
    def from_dict(cls, dikt) -> 'PoaPostData':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The poa_post_data of this PoaPostData.  # noqa: E501
        :rtype: PoaPostData
        """
        return util.deserialize_model(dikt, cls)

    @property
    def vcpu_cpu_map(self) -> List[PoaPostDataVcpuCpuMap]:
        """Gets the vcpu_cpu_map of this PoaPostData.


        :return: The vcpu_cpu_map of this PoaPostData.
        :rtype: List[PoaPostDataVcpuCpuMap]
        """
        return self._vcpu_cpu_map

    @vcpu_cpu_map.setter
    def vcpu_cpu_map(self, vcpu_cpu_map: List[PoaPostDataVcpuCpuMap]):
        """Sets the vcpu_cpu_map of this PoaPostData.


        :param vcpu_cpu_map: The vcpu_cpu_map of this PoaPostData.
        :type vcpu_cpu_map: List[PoaPostDataVcpuCpuMap]
        """

        self._vcpu_cpu_map = vcpu_cpu_map

    @property
    def node_set(self) -> List[str]:
        """Gets the node_set of this PoaPostData.


        :return: The node_set of this PoaPostData.
        :rtype: List[str]
        """
        return self._node_set

    @node_set.setter
    def node_set(self, node_set: List[str]):
        """Sets the node_set of this PoaPostData.


        :param node_set: The node_set of this PoaPostData.
        :type node_set: List[str]
        """

        self._node_set = node_set

    @property
    def bdf(self) -> List[str]:
        """Gets the bdf of this PoaPostData.


        :return: The bdf of this PoaPostData.
        :rtype: List[str]
        """
        return self._bdf

    @bdf.setter
    def bdf(self, bdf: List[str]):
        """Sets the bdf of this PoaPostData.


        :param bdf: The bdf of this PoaPostData.
        :type bdf: List[str]
        """

        self._bdf = bdf

    @property
    def keys(self) -> List[PoaPostDataKeys]:
        """Gets the keys of this PoaPostData.


        :return: The keys of this PoaPostData.
        :rtype: List[PoaPostDataKeys]
        """
        return self._keys

    @keys.setter
    def keys(self, keys: List[PoaPostDataKeys]):
        """Sets the keys of this PoaPostData.


        :param keys: The keys of this PoaPostData.
        :type keys: List[PoaPostDataKeys]
        """

        self._keys = keys



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
from typing import Tuple, List, Dict

from fabric_cf.actor.fim.fim_helper import FimHelper
from fim.slivers.attached_components import AttachedComponentsInfo, ComponentSliver, ComponentType
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.capacities_labels import Capacities, Labels
from fim.slivers.delegations import Delegations
from fim.slivers.instance_catalog import InstanceCatalog
from fim.slivers.interface_info import InterfaceSliver
from fim.slivers.network_node import NodeSliver, NodeType
from fim.slivers.network_service import NSLayer

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import BrokerException, ExceptionErrorCode
from fabric_cf.actor.core.kernel.reservation_states import ReservationOperation
from fabric_cf.actor.core.policy.inventory_for_type import InventoryForType
from fabric_cf.actor.core.util.id import ID


class NetworkNodeInventory(InventoryForType):
    @staticmethod
    def check_capacities(*, rid: ID, requested_capacities: Capacities, delegated: Delegations,
                         existing_reservations: List[ABCReservationMixin],
                         logger: logging.Logger) -> str:
        """
        Check if the requested capacities can be satisfied with the available capacities
        :param rid: reservation id of the reservation being served
        :param requested_capacities: Requested Capacities
        :param delegated: Delegated Capacities
        :param existing_reservations: Existing Reservations served by the same BQM node
        :param logger: logger
        :return: Delegation Id of the delegation which satisfies the request
        :raises: BrokerException in case the request cannot be satisfied
        """
        logger.debug(f"requested_capacities: {requested_capacities} for reservation# {rid}")

        delegation_id, delegated_capacity = FimHelper.get_delegations(delegations=delegated)

        # Remove allocated capacities to the reservations
        if existing_reservations is not None:
            for reservation in existing_reservations:
                if rid == reservation.get_reservation_id():
                    continue
                # For Active or Ticketed or Ticketing reservations; reduce the counts from available
                resource_sliver = InventoryForType.get_allocated_sliver(reservation=reservation)

                if resource_sliver is not None and isinstance(resource_sliver, NodeSliver):
                    logger.debug(
                        f"Excluding already assigned resources {resource_sliver.get_capacity_allocations()} to "
                        f"reservation# {reservation.get_reservation_id()}")
                    delegated_capacity = delegated_capacity - resource_sliver.get_capacity_allocations()

        # Compare the requested against available
        delegated_capacity = delegated_capacity - requested_capacities
        negative_fields = delegated_capacity.negative_fields()
        if len(negative_fields) > 0:
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"{negative_fields}")
        return delegation_id

    @staticmethod
    def __set_ips(*, req_ifs: InterfaceSliver, lab: Labels):
        if req_ifs.labels is not None and req_ifs.labels.ipv4 is not None:
            lab.ipv4 = req_ifs.labels.ipv4
            if req_ifs.labels.ipv4_subnet is not None:
                lab.ipv4_subnet = req_ifs.labels.ipv4_subnet
        if req_ifs.labels is not None and req_ifs.labels.ipv6 is not None:
            lab.ipv6 = req_ifs.labels.ipv6
            if req_ifs.labels.ipv6_subnet is not None:
                lab.ipv6_subnet = req_ifs.labels.ipv6_subnet
        return lab

    @staticmethod
    def __update_shared_nic_labels_and_capacities(*, available: ComponentSliver,
                                                  requested: ComponentSliver,
                                                  logger: logging.Logger) -> ComponentSliver:
        """
        Update the shared NIC Labels and Capacities. Assign the 1st available PCI address/bdf to the requested component
        Traverse the available component's labels to find the index for bdf assigned
        Using the found labels, assign BDF, MAC and VLAN address to the IFS on the Requested component
        In case of L2 service, also copy the requested IP address so it can be used by the AMHandler to configure the
        interface post VM creation
        :param available: Available Component
        :param requested: Requested Component
        :return updated requested component with VLAN, MAC and IP information
        """
        # Check labels
        delegation_id, delegated_label = FimHelper.get_delegations(
            delegations=available.get_label_delegations())

        if delegated_label.bdf is None or len(delegated_label.bdf) < 1:
            message = "No PCI devices available in the delegation"
            logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"{message}")

        # Find the VLAN from the BQM Component
        if available.network_service_info is None or \
                len(available.network_service_info.network_services) != 1:
            message = "Shared NIC Card must have one Network Service"
            logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.FAILURE,
                                  msg=f"{message}")

        ns_name = next(iter(available.network_service_info.network_services))
        ns = available.network_service_info.network_services[ns_name]

        if ns.interface_info is None or len(ns.interface_info.interfaces) != 1:
            message = "Shared NIC Card must have one Connection Point"
            logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.FAILURE,
                                  msg=f"{message}")

        ifs_name = next(iter(ns.interface_info.interfaces))
        ifs = ns.interface_info.interfaces[ifs_name]

        delegation_id, ifs_delegated_labels = FimHelper.get_delegations(delegations=ifs.get_label_delegations())

        assigned_bdf = delegated_label.bdf[0]
        assigned_numa = delegated_label.numa[0]

        # Updated the Requested component with VLAN, BDF, MAC
        req_ns_name = next(iter(requested.network_service_info.network_services))
        req_ns = requested.network_service_info.network_services[req_ns_name]
        req_ifs_name = next(iter(req_ns.interface_info.interfaces))
        req_ifs = req_ns.interface_info.interfaces[req_ifs_name]

        # Check if the requested component's VLAN exists in the delegated labels
        if req_ifs.labels and req_ifs.labels.vlan and \
                req_ifs.labels.vlan in ifs_delegated_labels.vlan:
            vlan_index = ifs_delegated_labels.vlan.index(req_ifs.labels.vlan)
            bdf_for_requested_vlan = ifs_delegated_labels.bdf[vlan_index]
            
            if bdf_for_requested_vlan in delegated_label.bdf:
                bdf_index = delegated_label.bdf.index(bdf_for_requested_vlan)
                assigned_bdf = bdf_for_requested_vlan
                assigned_numa = delegated_label.numa[bdf_index]

        # Assign the first PCI Id from the list of available PCI slots
        requested.label_allocations = Labels(bdf=assigned_bdf, numa=assigned_numa)

        # Find index of assigned BDF in the interface delegated labels
        assigned_index = ifs_delegated_labels.bdf.index(assigned_bdf)

        # Do not copy VLAN for OpenStack-vNIC
        if requested.get_model() == Constants.OPENSTACK_VNIC_MODEL:
            lab = Labels(bdf=ifs_delegated_labels.bdf[assigned_index], mac=ifs_delegated_labels.mac[assigned_index],
                         local_name=ifs_delegated_labels.local_name[assigned_index])
        else:
            lab = Labels(bdf=ifs_delegated_labels.bdf[assigned_index], mac=ifs_delegated_labels.mac[assigned_index],
                         vlan=ifs_delegated_labels.vlan[assigned_index],
                         local_name=ifs_delegated_labels.local_name[assigned_index])

        # For the Layer 2 copying the IP address to the label allocations
        # This is to be used by AM Handler to configure Network Interface
        if req_ns.layer == NSLayer.L2:
            lab = NetworkNodeInventory.__set_ips(req_ifs=req_ifs, lab=lab)

        req_ifs.set_label_allocations(lab=lab)

        logger.info(f"Assigned Interface Sliver: {req_ifs}")
        return requested

    @staticmethod
    def __update_smart_nic_labels_and_capacities(*, available: ComponentSliver,
                                                 requested: ComponentSliver,
                                                 logger: logging.Logger) -> ComponentSliver:
        """
        Update the IFS for the Smart NIC with VLAN, MAC and IP Address information
        This is to enable AM handler to configure network interfaces at VM creation.
        This is only done for Layer 2 services
        :param available: Available Component
        :param requested: Requested Component
        :return updated requested component with VLAN, MAC and IP information
        """

        # Find the VLAN from the BQM Component
        if available.network_service_info is None or \
                len(available.network_service_info.network_services) != 1:
            message = "Smart NIC Card must have at one Network Service"
            logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.FAILURE,
                                  msg=f"{message}")

        ns_name = next(iter(available.network_service_info.network_services))
        ns = available.network_service_info.network_services[ns_name]

        if ns.interface_info is None or len(ns.interface_info.interfaces) < 0:
            message = "Smart NIC Card must have at least one Connection Point"
            logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.FAILURE,
                                  msg=f"{message}")

        for ifs in ns.interface_info.interfaces.values():
            delegation_id, ifs_delegated_labels = FimHelper.get_delegations(
                delegations=ifs.get_label_delegations())

            for requested_ns in requested.network_service_info.network_services.values():
                if requested_ns.interface_info is not None and requested_ns.interface_info.interfaces is not None:
                    for requested_ifs in requested_ns.interface_info.interfaces.values():
                        if requested_ifs.labels.local_name == ifs_delegated_labels.local_name:
                            lab = Labels()
                            lab.mac = ifs_delegated_labels.mac
                            lab.local_name = ifs_delegated_labels.local_name

                            # Update the VLAN and IP address to be used for configuration at AM only for L2 services
                            # Information for L3 services is updated later after NetworkService has been ticketed
                            if requested_ns.layer == NSLayer.L2:
                                if requested_ifs.labels is not None and requested_ifs.labels.vlan is not None:
                                    lab.vlan = requested_ifs.labels.vlan

                                lab = NetworkNodeInventory.__set_ips(req_ifs=requested_ifs, lab=lab)

                            requested_ifs.set_label_allocations(lab=lab)
                        logger.info(f"Assigned Interface Sliver: {requested_ifs}")
        return requested

    @staticmethod
    def __check_component_labels_and_capacities(*, available: ComponentSliver, graph_id: str,
                                                requested: ComponentSliver, logger: logging.Logger,
                                                operation: ReservationOperation = ReservationOperation.Create) -> ComponentSliver:
        """
        Check if available component capacities, labels to match requested component
        :param available: available component
        :param graph_id: BQM graph id
        :param requested: requested component
        :param operation: operation
        :return: requested component annotated with properties in case of success, None otherwise
        """
        if requested.get_model() is not None and \
                requested.get_model() != available.get_model():
            return requested

        # Checking capacity for component
        delegation_id, delegated_capacity = FimHelper.get_delegations(
            delegations=available.get_capacity_delegations())

        # Delegated capacity would have been decremented already to exclude allocated shared NICs
        if delegated_capacity.unit < 1:
            message = f"Insufficient Capacities for component: {requested}"
            logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"{message}")

        requested.capacity_allocations = Capacities(unit=1)

        # Check labels
        delegation_id, delegated_label = FimHelper.get_delegations(
            delegations=available.get_label_delegations())

        if requested.get_type() == ComponentType.SharedNIC:
            requested = NetworkNodeInventory.__update_shared_nic_labels_and_capacities(
                available=available,
                requested=requested,
                logger=logger)
        else:
            requested.label_allocations = delegated_label
            if requested.get_type() == ComponentType.SmartNIC:
                requested = NetworkNodeInventory.__update_smart_nic_labels_and_capacities(
                    available=available,
                    requested=requested,
                    logger=logger)

        node_map = tuple([graph_id, available.node_id])
        requested.set_node_map(node_map=node_map)
        if requested.labels is None or operation == ReservationOperation.Create:
            requested.labels = Labels.update(lab=requested.get_label_allocations())

        return requested

    @staticmethod
    def __exclude_allocated_pci_device_from_shared_nic(*, shared: ComponentSliver, logger: logging.Logger,
                                                       allocated: ComponentSliver) -> Tuple[ComponentSliver, bool]:
        """
        For Shared NIC cards, exclude the already assigned PCI addresses from the available PCI addresses in
        BQM Component Sliver for the NIC Card
        @param shared: Available Shared NIC
        @param allocated: Allocated NIC
        @return Available NIC updated to exclude the Allocated PCI addresses and True/False indicating if Available
        Shared NIC has any available PCI addresses
        """

        if shared.get_type() != ComponentType.SharedNIC and allocated.get_type() != ComponentType.SharedNIC:
            raise BrokerException(error_code=ExceptionErrorCode.INVALID_ARGUMENT,
                                  msg=f"shared_nic: {shared} allocated_nic: {allocated}")

        # Reduce capacity for component
        delegation_id, delegated_capacity = FimHelper.get_delegations(
            delegations=shared.get_capacity_delegations())

        logger.debug(f"Allocated NIC: {allocated} labels: {allocated.get_labels()}")

        # Get the Allocated PCI address
        allocated_labels = allocated.get_labels()

        delegation_id, delegated_label = FimHelper.get_delegations(
            delegations=shared.get_label_delegations())

        # Remove allocated PCI address from delegations
        excluded_labels = []

        if isinstance(allocated_labels.bdf, list):
            excluded_labels = allocated_labels.bdf
        else:
            excluded_labels = [allocated_labels.bdf]

        exists = False
        for e in excluded_labels:
            if e in delegated_label.bdf:
                logger.debug(f"Excluding PCI device {e}")
                delegated_label.bdf.remove(e)
                exists = True

        # Exclude already allocated Shared NIC cards
        if exists:
            delegated_capacity -= allocated.get_capacity_allocations()

        return shared, (delegated_capacity.unit < 1)

    @staticmethod
    def __exclude_allocated_component(*, graph_node: NodeSliver, available: ComponentSliver,
                                      allocated: ComponentSliver, logger: logging.Logger):
        """
        Remove the allocated component from the candidate Node. For dedicated components, the whole component is removed,
        for Shared NIC, only the allocated PCI address is removed and the number of units is reduced by 1.
        If all the PCIs are allocated for a Shared NIC, the complete Shared NIC is removed

        @param graph_node candidate node identified to satisfy the reservation
        @param available available component
        @param allocated allocated component
        """
        exclude = True
        if allocated.get_type() == ComponentType.SharedNIC:
            available, exclude = NetworkNodeInventory.__exclude_allocated_pci_device_from_shared_nic(
                shared=available, allocated=allocated, logger=logger)
        if exclude:
            graph_node.attached_components_info.remove_device(name=available.get_name())

    @staticmethod
    def __exclude_components_for_existing_reservations(*, rid: ID, graph_node: NodeSliver, logger: logging.Logger,
                                                       existing_reservations: List[ABCReservationMixin],
                                                       operation: ReservationOperation = ReservationOperation.Create) -> NodeSliver:
        """
        Remove already assigned components to existing reservations from the candidate node
        @param rid reservation ID
        @param graph_node candidate node identified to satisfy the reservation
        @param existing_reservations Existing Ticketed Reservations
        @return Return the updated candidate node
        """
        for reservation in existing_reservations:
            # Requested reservation should be skipped only when new i.e. not ticketed
            if rid == reservation.get_reservation_id() and \
                    (operation == ReservationOperation.Extend or not reservation.is_ticketed()):
                continue
            # For Active or Ticketed or Ticketing reservations; reduce the counts from available
            allocated_sliver = InventoryForType.get_allocated_sliver(reservation=reservation)

            if reservation.is_extending_ticket() and reservation.get_requested_resources() is not None and \
                    reservation.get_requested_resources().get_sliver() is not None:
                allocated_sliver = reservation.get_requested_resources().get_sliver()

            if allocated_sliver is None or not isinstance(allocated_sliver, NodeSliver) or \
                    allocated_sliver.attached_components_info is None:
                continue

            for allocated in allocated_sliver.attached_components_info.devices.values():
                allocated_node_map = allocated.get_node_map()

                if allocated_node_map is None:
                    continue

                resource_type = allocated.get_type()

                logger.debug(f"Already allocated components {allocated} of resource_type "
                             f"{resource_type} to reservation# {reservation.get_reservation_id()}")

                for av in graph_node.attached_components_info.devices.values():
                    if av.node_id == allocated_node_map[1]:
                        NetworkNodeInventory.__exclude_allocated_component(graph_node=graph_node,
                                                                           available=av,
                                                                           allocated=allocated,
                                                                           logger=logger)
                        break
        return graph_node

    @staticmethod
    def check_components(*, rid: ID, requested_components: AttachedComponentsInfo, graph_id: str,
                         graph_node: NodeSliver, existing_reservations: List[ABCReservationMixin],
                         existing_components: Dict[str, List[str]], logger: logging.Logger,
                         operation: ReservationOperation = ReservationOperation.Create) -> AttachedComponentsInfo:
        """
        Check if the requested capacities can be satisfied with the available capacities
        :param rid: reservation id of the reservation being served
        :param requested_components: Requested components
        :param graph_id: BQM graph id
        :param graph_node: BQM graph node identified to serve the reservation
        :param existing_reservations: Existing Reservations served by the same BQM node
        :param existing_components: Existing components
        :param operation: Flag indicating if this is create or modify
        :param logger: logger
        :return: Components updated with the corresponding BQM node ids
        :raises: BrokerException in case the request cannot be satisfied
        """
        logger.debug(f"Available on {graph_node.node_id} components: {graph_node.attached_components_info.devices.keys()}")

        NetworkNodeInventory.__exclude_components_for_existing_reservations(rid=rid, graph_node=graph_node,
                                                                            existing_reservations=existing_reservations,
                                                                            operation=operation, logger=logger)

        logger.debug(f"Excluding components connected to Network Services: {existing_components}")

        if existing_components and len(existing_components):
            comps_to_remove = []
            for av in graph_node.attached_components_info.devices.values():
                # Skip if not in allocated comps attached to Network Services
                if av.node_id not in existing_components.keys():
                    continue
                exclude = True
                if av.get_type() == ComponentType.SharedNIC:
                    bdfs = existing_components.get(av.node_id)
                    allocated_component = ComponentSliver()
                    allocated_component.set_type(ComponentType.SharedNIC)
                    allocated_component.set_name(resource_name=av.get_name())
                    allocated_component.set_capacity_allocations(cap=Capacities(unit=len(bdfs)))
                    allocated_component.set_labels(Labels(bdf=bdfs))
                    logger.debug(f"Excluding Shared NICs connected to Network Services: {allocated_component}")
                    av, exclude = NetworkNodeInventory.__exclude_allocated_pci_device_from_shared_nic(shared=av,
                                                                                                      allocated=allocated_component,
                                                                                                      logger=logger)
                if exclude:
                    comps_to_remove.append(av)

            for c in comps_to_remove:
                logger.debug(f"Excluding component: {c.get_name()}")
                graph_node.attached_components_info.remove_device(name=c.get_name())

        logger.debug(f"requested_components: {requested_components.devices.values()} for reservation# {rid}")
        for name, requested_component in requested_components.devices.items():
            if operation == ReservationOperation.Modify and requested_component.get_node_map() is not None:
                logger.debug(f"Modify: Ignoring Allocated component: {requested_component}")
                continue

            if operation == ReservationOperation.Extend and requested_component.get_node_map() is not None:
                bqm_id, node_id = requested_component.get_node_map()

                if requested_component.get_type() == ComponentType.SharedNIC:
                    allocated_bdfs = existing_components.get(node_id)
                    if allocated_bdfs and requested_component.labels and requested_component.labels.bdf:
                        bdfs = requested_component.labels.bdf
                        if isinstance(requested_component.labels.bdf, str):
                            bdfs = [requested_component.labels.bdf]

                        logger.debug(f"Allocated BDFs: {allocated_bdfs}")
                        for x in bdfs:
                            if x in allocated_bdfs:
                                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                                      msg=f"Renew failed: Component of type: {requested_component.get_model()} with PCI Address: {x}"
                                                          f"already in use by another reservation for node: {graph_node.node_id}")
                else:
                    if node_id in existing_components.keys():
                        raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                              msg=f"Renew failed: Component of type: {requested_component.get_model()} "
                                                  f"already in use by another reservation for node: {graph_node.node_id}")

                logger.debug(f"Renew: Component {requested_component} still available")
                continue

            logger.debug(f"Create: Allocating component: {requested_component}")
            resource_type = requested_component.get_type()
            resource_model = requested_component.get_model()
            if resource_type == ComponentType.Storage:
                requested_component.capacity_allocations = Capacities(unit=1)
                requested_component.label_allocations = Labels()
                requested_component.label_allocations = Labels.update(lab=requested_component.get_labels())
                continue
            available_components = graph_node.attached_components_info.get_devices_by_type(resource_type=resource_type)
            logger.debug(f"Available components of type: {resource_type} after excluding "
                              f"allocated components: {available_components}")

            if available_components is None or len(available_components) == 0:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Component of type: {resource_model} not available in "
                                          f"graph node: {graph_node.node_id}")

            for component in available_components:
                # check model matches the requested model
                requested_component = NetworkNodeInventory.__check_component_labels_and_capacities(
                    available=component, graph_id=graph_id,
                    requested=requested_component,
                    operation=operation, logger=logger)

                if requested_component.get_node_map() is not None:
                    logger.info(f"Assigning {component.node_id} to component# "
                                     f"{requested_component} in reservation# {rid} ")

                    # Remove the component from available components as it is assigned
                    NetworkNodeInventory.__exclude_allocated_component(graph_node=graph_node, available=component,
                                                       allocated=requested_component, logger=logger)
                    break

            if requested_component.get_node_map() is None:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Component of type: {resource_model} not available in "
                                          f"graph node: {graph_node.node_id}")

        return requested_components

    def __allocate_p4_switch(self, *, rid: ID, requested_sliver: NodeSliver, graph_id: str, graph_node: NodeSliver,
                             existing_reservations: List[ABCReservationMixin], existing_components: Dict[str, List[str]],
                             operation: ReservationOperation = ReservationOperation.Create) -> Tuple[str, BaseSliver]:
        """
        Allocate an extending or ticketing reservation for a P4 switch

        :param rid: reservation id of the reservation to be allocated
        :param requested_sliver: requested sliver
        :param graph_id: BQM graph id
        :param graph_node: BQM graph node identified to serve the reservation
        :param existing_components: Existing Components
        :param existing_reservations: Existing Reservations served by the same BQM node
        :param operation: Indicates if this is create or modify

        :return: Tuple of Delegation Id and the Requested Sliver annotated with BQM Node Id and other properties
        :raises: BrokerException in case the request cannot be satisfied
        """
        delegation_id = None

        if operation == ReservationOperation.Create:
            # In case of modify, directly get delegation_id
            if len(graph_node.get_capacity_delegations().get_delegation_ids()) > 0:
                delegation_id = next(iter(graph_node.get_capacity_delegations().get_delegation_ids()))

        # Handle allocation to account for leaked Network Services
        for n in existing_components.keys():
            if n in graph_node.node_id:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Node of type: {graph_node.get_type()} not available on site: "
                                          f"{graph_node.get_site()}, already in use by another reservation")

        # For create, we need to allocate the P4
        requested_capacities = requested_sliver.get_capacities()

        # Check if Capacities can be satisfied
        delegation_id = self.check_capacities(rid=rid, requested_capacities=requested_capacities,
                                              delegated=graph_node.get_capacity_delegations(),
                                              existing_reservations=existing_reservations, logger=self.logger)
        requested_sliver.capacity_allocations = Capacities()
        requested_sliver.capacity_allocations = Capacities.update(lab=requested_capacities)
        requested_sliver.label_allocations = Labels(local_name=graph_node.get_name())

        requested_sliver.set_node_map(node_map=(graph_id, graph_node.node_id))
        requested_sliver.management_ip = graph_node.management_ip

        self.logger.info(f"Reservation# {rid} is being served by delegation# {delegation_id} "
                         f"node# [{graph_id}/{graph_node.node_id}]")

        return delegation_id, requested_sliver

    def allocate(self, *, rid: ID, requested_sliver: BaseSliver, graph_id: str, graph_node: BaseSliver,
                 existing_reservations: List[ABCReservationMixin], existing_components: Dict[str, List[str]],
                 operation: ReservationOperation = ReservationOperation.Create) -> Tuple[str, BaseSliver]:
        """
        Allocate an extending or ticketing reservation
        :param rid: reservation id of the reservation to be allocated
        :param requested_sliver: requested sliver
        :param graph_id: BQM graph id
        :param graph_node: BQM graph node identified to serve the reservation
        :param existing_components: Existing Components
        :param existing_reservations: Existing Reservations served by the same BQM node
        :param operation: Indicates if this is create or modify
        :return: Tuple of Delegation Id and the Requested Sliver annotated with BQM Node Id and other properties
        :raises: BrokerException in case the request cannot be satisfied
        """
        if graph_node.get_capacity_delegations() is None or rid is None:
            raise BrokerException(error_code=ExceptionErrorCode.INVALID_ARGUMENT,
                                  msg=f"capacity_delegations is missing or reservation is None")

        if not isinstance(requested_sliver, NodeSliver):
            raise BrokerException(error_code=ExceptionErrorCode.INVALID_ARGUMENT,
                                  msg=f"resource type: {requested_sliver.get_type()}")

        if not isinstance(graph_node, NodeSliver):
            raise BrokerException(error_code=ExceptionErrorCode.INVALID_ARGUMENT,
                                  msg=f"resource type: {graph_node.get_type()}")

        if requested_sliver.get_type() not in [NodeType.VM, NodeType.Switch]:
            raise BrokerException(error_code=ExceptionErrorCode.INVALID_ARGUMENT,
                                  msg=f"Unsupported resource type: {graph_node.get_type()}")

        if requested_sliver.get_type() == NodeType.Switch:
            return self.__allocate_p4_switch(rid=rid, requested_sliver=requested_sliver, graph_id=graph_id,
                                             graph_node=graph_node, existing_reservations=existing_reservations,
                                             existing_components=existing_components, operation=operation)

        delegation_id = None
        requested_capacities = None
        # For create, we need to allocate the VM
        if operation == ReservationOperation.Create:
            # Always use requested capacities to be mapped from flavor i.e. capacity hints
            requested_capacity_hints = requested_sliver.get_capacity_hints()
            catalog = InstanceCatalog()
            requested_capacities = catalog.get_instance_capacities(instance_type=requested_capacity_hints.instance_type)
        else:
            requested_capacities = requested_sliver.get_capacity_allocations()
            # In case of modify, directly get delegation_id
            if len(graph_node.get_capacity_delegations().get_delegation_ids()) > 0:
                delegation_id = next(iter(graph_node.get_capacity_delegations().get_delegation_ids()))

        # Check if Capacities can be satisfied
        delegation_id = self.check_capacities(rid=rid, requested_capacities=requested_capacities,
                                              delegated=graph_node.get_capacity_delegations(),
                                              existing_reservations=existing_reservations, logger=self.logger)

        # Check if Components can be allocated
        if requested_sliver.attached_components_info is not None:
            requested_sliver.attached_components_info = self.check_components(
                rid=rid,
                requested_components=requested_sliver.attached_components_info,
                graph_id=graph_id,
                graph_node=graph_node,
                existing_reservations=existing_reservations,
                existing_components=existing_components,
                operation=operation,
                logger=self.logger)

        # Do this only for create
        if operation == ReservationOperation.Create:
            requested_sliver.capacity_allocations = Capacities()
            requested_sliver.capacity_allocations = Capacities.update(lab=requested_capacities)
            requested_sliver.label_allocations = Labels(instance_parent=graph_node.get_name())

            requested_sliver.set_node_map(node_map=(graph_id, graph_node.node_id))

        self.logger.info(f"Reservation# {rid} is being served by delegation# {delegation_id} "
                         f"node# [{graph_id}/{graph_node.node_id}]")

        return delegation_id, requested_sliver

    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        pass



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_tick import ABCTick


class ABCTicker(ABC):
    """
    ITicker defines the interface for the internal clock of a container. A container measures time in units
    of cycles. A cycle can have a variable length, which is specified at boot time.
    Container clocks can optionally be configured with an offset to be applied when calculating the current cycle.
    For more details about cycles see Time within the container can advance automatically, based on the internal
    computer clock, or manually, based on calls to the container clock's tick method.

    A clock implementation must support at least one mode of execution (manual or automatic), but it need not
    support both.

    A clock implementation maintains a collection of subscribers, which will be notified about the passage of time.
    Subscriptions are dynamic and can be changed at run time.

    Abstractly, each subscriber will receive a notification about the advancement of the container clock.
    Some implementations may guarantee that if an object has received a notification for cycle x, the next
    notification will be for cycle x+1. However, in general, no such guarantees are provided. As a matter of fact,
    code executing in real time will often "skip" cycles. Therefore, subscribed object, in general, must be able
    to handle correctly cases in which one or more clock cycle notifications are omitted.
    """

    @abstractmethod
    def tick(self):
        """
        Advances the clock with one cycle. This operation is only
        applicable for manual clocks. This is a blocking operation.
        """

    @abstractmethod
    def start(self):
        """
        Starts the clock.
        """

    @abstractmethod
    def stop(self):
        """
        Stops/pauses the clock.

        @throws Exception in case of error
        """

    @abstractmethod
    def add_tickable(self, *, tickable: ABCTick):
        """
        Adds an object to the list of objects subscribed to receive tick
        events.

        @param tickable object to add
        """

    @abstractmethod
    def remove_tickable(self, *, tickable: ABCTick):
        """
        Removes an object from the list of object subscribed to receive
        tick events.

        @param tickable object to remove
        """

    @abstractmethod
    def clear(self):
        """
        Stops the clock and unregisters all registered items.

        @throws Exception in case of error
        """

    @abstractmethod
    def get_current_cycle(self) -> int:
        """
        Returns the current clock cycle.

        @return current clock cycle
        """

    @abstractmethod
    def set_current_cycle(self, *, cycle: int):
        """
        Sets the current cycle. The first cycle to be generated by the
        clock will be cycle+1. This method can called only if
        initialize has not yet been called.

        @param cycle current cycle
        """

    @abstractmethod
    def set_beginning_of_time(self, *, value: int):
        """
        Sets the clock offset (relative to current time). Clock offset
        is measured in milliseconds. Clock offset is used to calculate the
        current clock cycle by subtracting it from current time and dividing by
        the length of a cycle. This method can called only if
        initialize has not yet been called.

        @param value clock offset in milliseconds
        """

    @abstractmethod
    def get_beginning_of_time(self) -> int:
        """
        Returns the clock offset. Clock offset is measured in
        milliseconds.

        @return clock offset in milliseconds
        """

    @abstractmethod
    def get_cycle_millis(self) -> int:
        """
        Returns the cycle length. Cycle length is measured in
        milliseconds.

        @return cycle length in milliseconds
        """

    @abstractmethod
    def set_cycle_millis(self, *, cycle_millis: int):
        """
        Sets the cycle length. Cycle length is measured in milliseconds.
        This method can called only if initialize has not yet been
        called.

        @param cycleMillis cycle length in milliseconds
        """

    @abstractmethod
    def is_manual(self) -> bool:
        """
        Checks if the clock implementation is manual, i.e., time
        advances manually as a result of calls to tick

        @return true if the clock is manual, false otherwise.
        """

    @abstractmethod
    def set_manual(self, *, value: bool):
        """
        Sets the manual clock flag. This method may throw a
        Exception if the clock does not support the
        particular execution mode. This method can called only if
        initialize has not yet been called.

        @param value value for the manual flag
        """



import time
import unittest
from datetime import datetime

from fabric_cf.actor.fim.fim_helper import FimHelper

from fabric_cf.actor.core.kernel.reservation_states import ReservationStates

from fabric_cf.actor.core.plugins.db.actor_database import ActorDatabase


from fim.graph.abc_property_graph import ABCPropertyGraph, GraphFormat
from fim.graph.abc_property_graph_constants import ABCPropertyGraphConstants
from fim.graph.neo4j_property_graph import Neo4jGraphImporter, Neo4jPropertyGraph
from fim.graph.resources.neo4j_arm import Neo4jARMGraph
from fim.graph.resources.neo4j_cbm import Neo4jCBMFactory, Neo4jCBMGraph
from fim.slivers.path_info import Path
from fim.user import NodeType, ERO, ExperimentTopology, Capacities, ComponentModelType, ServiceType

from fabric_cf.actor.fim.plugins.broker.aggregate_bqm_plugin import AggregatedBQMPlugin
from fabric_cf.actor.test.base_test_case import BaseTestCase

"""
Test of an ABQM plugin
"""
import yaml
neo4j = None
with open("./config/config.test.yaml", 'r') as stream:
    try:
        config_dict = yaml.safe_load(stream)
        neo4j = config_dict["neo4j"]
    except yaml.YAMLError as exc:
        print(exc)


class ABQM_Test(BaseTestCase, unittest.TestCase):
    def get_clean_database(self) -> ActorDatabase:
        db = self.get_actor_database()
        db.set_actor_name(name=self.actor_name)
        db.set_reset_state(state=True)
        db.initialize()
        return db

    n4j_imp = Neo4jGraphImporter(url=neo4j["url"], user=neo4j["user"],
                                 pswd=neo4j["pass"],
                                 import_host_dir=neo4j["import_host_dir"],
                                 import_dir=neo4j["import_dir"])

    def test_abqm(self):
        self.n4j_imp.delete_all_graphs()
        # these are produced by substrate tests
        site_ads = ['../../../neo4j/Network-dev.graphml',
                    '../../../neo4j/LBNL.graphml',
                    '../../../neo4j/RENC.graphml',
                    '../../../neo4j/UKY.graphml']

        cbm = Neo4jCBMGraph(importer=self.n4j_imp)

        adm_ids = dict()

        for ad in site_ads:
            plain_neo4j = self.n4j_imp.import_graph_from_file_direct(graph_file=ad)
            print(f"Validating ARM graph {ad}")
            plain_neo4j.validate_graph()

            site_arm = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=plain_neo4j.graph_id,
                                                              importer=self.n4j_imp))
            # generate a dict of ADMs from site graph ARM
            site_adms = site_arm.generate_adms()
            print('ADMS' + str(site_adms.keys()))

            # desired ADM is under 'primary'
            site_adm = site_adms['primary']
            cbm.merge_adm(adm=site_adm)

            print('Deleting ADM and ARM graphs')
            for adm in site_adms.values():
                adm_ids[ad] = adm.graph_id
                adm.delete_graph()
            site_arm.delete_graph()

        cbm.validate_graph()
        print('CBM ID is ' + cbm.graph_id)

        cbm_graph_id = cbm.graph_id
        # turn on debug so we can test formation of ABQM without querying
        # actor for reservations
        AggregatedBQMPlugin.DEBUG_FLAG = True

        n4j_pg = Neo4jPropertyGraph(graph_id=cbm_graph_id, importer=self.n4j_imp)

        cbm = Neo4jCBMFactory.create(n4j_pg)

        plugin = AggregatedBQMPlugin(actor=None, logger=None)

        abqm = plugin.plug_produce_bqm(cbm=cbm, query_level=1)

        abqm.validate_graph()

        abqm_string = abqm.serialize_graph()

        print('Writing ABQM to abqm.graphml')
        with open('abqm.graphml', 'w') as f:
            f.write(abqm_string)

        abqm_json = abqm.serialize_graph(format=GraphFormat.JSON_NODELINK)
        print('Writing ABQM to abqm.json')
        with open('abqm.json', 'w') as f:
            f.write(abqm_json)

        abqm_level2 = plugin.plug_produce_bqm(cbm=cbm, query_level=2)

        abqm_level2.validate_graph()

        abqm_level2_string = abqm_level2.serialize_graph()

        print('Writing ABQM to abqm_level2.graphml')
        with open('abqm_level2.graphml', 'w') as f:
            f.write(abqm_level2_string)

        abqm_level2_json = abqm.serialize_graph(format=GraphFormat.JSON_NODELINK)
        print('Writing ABQM to abqm_level2.json')
        with open('abqm_level2.json', 'w') as f:
            f.write(abqm_level2_json)

        print('Writing CBM to cbm.graphml')
        cbm_string = cbm.serialize_graph()
        with open('cbm.graphml', 'w') as f:
            f.write(cbm_string)

        plain_cbm = self.n4j_imp.import_graph_from_string_direct(graph_string=abqm_level2_string)
        temp = Neo4jCBMFactory.create(Neo4jPropertyGraph(graph_id=plain_cbm.graph_id,
                                                         importer=self.n4j_imp))

        site_node_ids = {}
        for s in temp.get_all_nodes_by_class_and_type(label=ABCPropertyGraph.CLASS_CompositeNode,
                                                      ntype=str(NodeType.Server)):
            labels, props = temp.get_node_properties(node_id=s)
            site_node_ids[props.get('Site')] = s

        ns_node_ids = {}
        for s in temp.get_all_network_service_nodes():
            labels, props = temp.get_node_properties(node_id=s)
            ns_node_ids[props.get('Name')] = s

        path = temp.get_nodes_on_path_with_hops(node_a=site_node_ids['UKY'], node_z=site_node_ids['LBNL'],
                                                hops=[ns_node_ids['RENC_ns']])

        assert(len(path) != 0)
        from fim.user.topology import AdvertizedTopology
        substrate = AdvertizedTopology()
        substrate.load(graph_string=abqm_level2_string)

        uky_node_id = substrate.sites.get("UKY").node_id
        lbnl_node_id = substrate.sites.get("LBNL").node_id
        renc_ns_node_id = substrate.network_services.get("RENC_ns").node_id

        path = substrate.graph_model.get_nodes_on_path_with_hops(node_a=uky_node_id, node_z=lbnl_node_id,
                                                                 hops=[renc_ns_node_id])

        assert(len(path) != 0)

        self.n4j_imp.delete_all_graphs()

    def test_cbm(self):
        self.n4j_imp.delete_all_graphs()
        # these are produced by substrate tests
        cbm = '../../../neo4j/abqm-l2-cbm.graphml'

        plain_cbm = self.n4j_imp.import_graph_from_file_direct(graph_file=cbm)
        cbm = Neo4jCBMFactory.create(Neo4jPropertyGraph(graph_id=plain_cbm.graph_id,
                                     importer=self.n4j_imp))
        cbm.validate_graph()

        print('CBM ID is ' + cbm.graph_id)

        cbm_graph_id = cbm.graph_id
        # turn on debug so we can test formation of ABQM without querying
        # actor for reservations
        AggregatedBQMPlugin.DEBUG_FLAG = True

        n4j_pg = Neo4jPropertyGraph(graph_id=cbm_graph_id, importer=self.n4j_imp)

        cbm = Neo4jCBMFactory.create(n4j_pg)

        plugin = AggregatedBQMPlugin(actor=None, logger=None)

        abqm = plugin.plug_produce_bqm(cbm=cbm, query_level=1)

        abqm.validate_graph()

        abqm_string = abqm.serialize_graph()

        print('Writing ABQM to abqm-from-cbm.graphml')
        with open('abqm-from-cbm.graphml', 'w') as f:
            f.write(abqm_string)

        abqm2 = plugin.plug_produce_bqm(cbm=cbm, query_level=2)

        abqm2.validate_graph()

        abqm2_string = abqm2.serialize_graph()

        print('Writing ABQM to abqm2-from-cbm.graphml')
        with open('abqm2-from-cbm.graphml', 'w') as f:
            f.write(abqm2_string)

        self.n4j_imp.delete_all_graphs()

    def test_load_abqm_level_zero(self):
        self.n4j_imp.delete_all_graphs()
        # these are produced by substrate tests
        cbm = 'cbm-dev.graphml'

        plain_cbm = self.n4j_imp.import_graph_from_file_direct(graph_file=cbm)
        cbm = Neo4jCBMFactory.create(Neo4jPropertyGraph(graph_id=plain_cbm.graph_id,
                                     importer=self.n4j_imp))
        cbm.validate_graph()

        print('CBM ID is ' + cbm.graph_id)

        cbm_graph_id = cbm.graph_id
        # turn on debug so we can test formation of ABQM without querying
        # actor for reservations
        AggregatedBQMPlugin.DEBUG_FLAG = True

        n4j_pg = Neo4jPropertyGraph(graph_id=cbm_graph_id, importer=self.n4j_imp)

        cbm = Neo4jCBMFactory.create(n4j_pg)

        plugin = AggregatedBQMPlugin(actor=None, logger=None)

        abqm = plugin.plug_produce_bqm(cbm=cbm, query_level=2)

        abqm.validate_graph()

        abqm_string = abqm.serialize_graph()

        cbm.delete_graph()

        plain_abqm = self.n4j_imp.import_graph_from_string_direct(graph_string=abqm_string)
        plain_abqm.validate_graph()

        print('Writing ABQM to abqm-from-cbm.graphml')
        with open('abqm-from-cbm.graphml', 'w') as f:
            f.write(abqm_string)

        from fim.user.topology import AdvertizedTopology
        substrate = AdvertizedTopology()
        substrate.load(graph_string=abqm_string)
        print(substrate)

        self.n4j_imp.delete_all_graphs()

    def test_ero_find_paths(self):
        cbm_graph_id = "162bf53f-85c5-498f-bc6f-d7d3109263a8"
        n4j_pg = Neo4jCBMGraph(graph_id=cbm_graph_id, importer=self.n4j_imp)

        hop = "WASH"
        t = ExperimentTopology()
        n1 = t.add_node(name='n1', site='HAWI')
        n2 = t.add_node(name='n2', site='CERN')
        cap = Capacities(core=2, ram=8, disk=100)
        n1.set_properties(capacities=cap, image_type='qcow2', image_ref='default_centos_8')
        n2.set_properties(capacities=cap, image_type='qcow2', image_ref='default_centos_8')

        n1.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name='n1-nic1')

        n2.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name='n2-nic1')

        ns = t.add_network_service(name='bridge1', nstype=ServiceType.L2PTP, interfaces=[n1.interface_list[0],
                                                                                         n2.interface_list[0]])

        hops = [n1.site, hop, n2.site]
        path = Path()
        path.set_symmetric(hops)
        ero = ERO()
        ero.set(payload=path)
        ns.ero = ero

        type, path = ns.ero.get()
        path_list = path.get()[0]

        source = n4j_pg.get_matching_nodes_with_components(label=ABCPropertyGraphConstants.CLASS_NetworkNode,
                                                           props={ABCPropertyGraphConstants.PROP_SITE: path_list[0]})

        dest = n4j_pg.get_matching_nodes_with_components(label=ABCPropertyGraphConstants.CLASS_NetworkNode,
                                                         props={ABCPropertyGraphConstants.PROP_SITE: path_list[-1]})
        hops = []
        for h in path_list:
            hop = n4j_pg.get_matching_nodes_with_components(label=ABCPropertyGraphConstants.CLASS_NetworkService,
                                                            props={ABCPropertyGraphConstants.PROP_NAME: f"{h}_ns"})
            hops.append(hop[0])

        print(f"HOPS: {hops}")
        final_path = []

        db = self.get_actor_database()


        #shortest_path = n4j_pg.get_nodes_on_shortest_path(node_a=source[0], node_z=dest[0])
        #print(f"Shortest Path: {shortest_path}")

        paths = n4j_pg.get_nodes_on_path_with_hops(node_a=source[0], node_z=dest[0], hops=hops)
        sorted_paths = sorted(paths, key=len)
        print(f"Number of paths: {len(sorted_paths)}")
        for sorted_path in sorted_paths:
            found = True
            links = []
            for item in sorted_path:
                _, props = n4j_pg.get_node_properties(node_id=item)
                #print(props)
                if item.startswith('link:'):
                    links.append(item)
                if props.get("Class") == "NetworkService":
                    final_path.append(f'{props.get("Name")}:{props.get("NodeID")}')

            print(f"Links in path: {links}")
            for l in links:
                link_sliver = n4j_pg.build_deep_link_sliver(node_id=l)
                print(f"""
                Link Info:
                  Node ID       : {link_sliver.node_id}
                  Type          : {link_sliver.get_type()}
                  Allowed Caps  : {link_sliver.capacity_allocations}
                  Total Caps    : {link_sliver.capacities}
                """)

                #if link_sliver.get_type() == LinkType.L1Path:
                #    print(link_sliver)
                #    break
            if found:
                break

        print(f"FINAL - Path: {final_path}")

        '''
        print("=============================================================================================")
        print(f"Number of paths: {len(sorted_paths)}")
        paths = n4j_pg.get_nodes_on_path_with_hops(node_a=dest[0], node_z=source[0], hops=hops)
        sorted_paths = sorted(paths, key=len)
        for sorted_path in sorted_paths:
            links = [item for item in sorted_path if item.startswith('link:')]
            print(f"Links in path: {links}")
            for l in links:
                _, props = n4j_pg.get_node_properties(node_id=l)
                print(props)
                if props['Type'] != 'L2Path':
                    break
            break
        '''